#include "systems.h"
#include "system_02.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
#ifndef __DDEM_INDEX_TRANSLATOR_DEF
#define __DDEM_INDEX_TRANSLATOR_DEF

#define Z(I,J) z[(J)*m + (I)]
/* Z(I,J) corresponds to y_I(alpha_J) if  0<= J < nu */
/* Z(I,J) corresponds to y'_I(alpha_J) if nu <= J < nu + omega */

#define TJAC(I) userProvidedJac->tJac[(I)]
#define YJAC(I,J) userProvidedJac->yJac[(I)*m + (J)]
#define ZJAC(I,J,K) userProvidedJac->zJac[(I)*(m*(nu+omega)) + (K)*m+(J)]
/* par F_I / par Z(J,K) */
#define PJAC(I,J) userProvidedJac->pJac[(I)*l + (J)]

#define TTJAC(I) userProvidedJac->ttJac[(I)]
#define TPJAC(I,J) userProvidedJac->tpJac[(I)*l + (J)]

#endif // __DDEM_INDEX_TRANSLATOR_DEF
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void	f_02(int state, double t, int m, double* y,
										  int nu, int omega, double* z,
											int l, double* p,
											double* f,
                      ddemUserProvidedJac* userProvidedJac,
											void* comm)
{
	// fIndices = 0, 1, 2, ..., (nVariables - 1) 
	// yIndices = 0, 1, 2, ..., (nVariables - 1) 
	// alphaIndices = 0, 1, 2, ..., (nu + omega - 1) 
	// pIndices = 0, 1, 2, ..., (nParameters - 1) 
	//*******************************************
	// set f[fIndices]
  double temp = y[0] * y[0] /( y[0] * y[0] + 1.0);
	f[0] = y[0] * (1.0 - Z(0,0) - p[1] * Z(0,1)) - y[1] * temp;
  f[1] = y[1] * (temp - p[2]);
	if(userProvidedJac->needed)
	{
		userProvidedJac->tIndependent = DDEM_TRUE;
		// or set TJAC(fIndices)
		// TJAC(0) = ;

		//userProvidedJac->yIndependent = DDEM_TRUE;
		// or set YJAC(fIndices, yIndices)
		double par_temp_par_y0 = 2.0 * y[0] / (( y[0] * y[0] + 1.0) * ( y[0] * y[0] + 1.0));
		YJAC(0,0) = (1.0 - Z(0,0) - p[1] * Z(0,1)) - y[1] * par_temp_par_y0;
		YJAC(0,1) = - temp;
		YJAC(1,0) = y[1] * par_temp_par_y0;
		YJAC(1,1) = temp - p[2];
		// YJAC(0,0) = ;

		// set ZJAC(fIndices, yIndices, alphaIndices)
		ZJAC(0,0,0) = - y[0];
		ZJAC(0,0,1) = - y[0] * p[1];
		ZJAC(0,1,0) = 0.0;
		ZJAC(0,1,1) = 0.0;

		ZJAC(1,0,0) = 0.0;
		ZJAC(1,0,1) = 0.0;
		ZJAC(1,1,0) = 0.0;
		ZJAC(1,1,1) = 0.0;

		//userProvidedJac->pIndependent = DDEM_TRUE;
		// or set PJAC(fIndices, pIndices)
		PJAC(0,0) = 0.0;
		PJAC(0,1) = - y[0] * Z(0,1);
		PJAC(0,2) = 0.0;
		PJAC(0,3) = 0.0; PJAC(0,4) = 0.0; PJAC(0,5) = 0.0; PJAC(0,6) = 0.0;

		PJAC(1,0) = 0.0;
		PJAC(1,1) = 0.0;
		PJAC(1,2) = - y[1];
		PJAC(1,3) = 0.0; PJAC(1,4) = 0.0; PJAC(1,5) = 0.0; PJAC(1,6) = 0.0;
	}
}
/*---------------------------------------------*/
void historyFunc_02(int historySegmentIndex,
											 double t,
                       int l, double* p,
                       int m, double* h,
                       ddemUserProvidedJac* userProvidedJac,
											 void* comm)
{
	// hIndices = 0, 1, 2, ..., (nVariables - 1) 
	// pIndices = 0, 1, 2, ..., (nParameters - 1) 
	//*******************************************
	// set h[hIndices]
  h[0] = p[3] + p[4] * t;
	h[1] = p[5] + p[6] * t;
	if(userProvidedJac->needed)
	{
		TJAC(0) = p[4];
		TJAC(1) = p[6];

		//userProvidedJac->pIndependent = DDEM_TRUE;
		// or set PJAC(hIndices, pIndices)
		PJAC(0,0) = 0.0; PJAC(0,1) = 0.0; PJAC(0,2) = 0.0;
		PJAC(0,3) = 1.0;
		PJAC(0,4) = t;
		PJAC(0,5) = 0.0; PJAC(0,6) = 0.0;

		PJAC(1,0) = 0.0; PJAC(1,1) = 0.0; PJAC(1,2) = 0.0;
		PJAC(1,3) = 0.0; PJAC(1,4) = 0.0;
		PJAC(1,5) = 1.0;
		PJAC(1,6) = t;

		if(userProvidedJac->needed2nd)
		{
			userProvidedJac->ttIndependent = DDEM_TRUE;
			// or set TTJAC(hIndices)
			// TTJAC(0) = ;
			
			userProvidedJac->tpIndependent = DDEM_TRUE;
			// or set TPJAC(hIndices, pIndices)
			// TPJAC(0,0) = ;
			TPJAC(0,0) = 0.0; TPJAC(0,1) = 0.0; TPJAC(0,2) = 0.0;
			TPJAC(0,3) = 0.0;
			TPJAC(0,4) = 1.0;
			TPJAC(0,5) = 0.0; TPJAC(0,6) = 0.0;

			TPJAC(1,0) = 0.0; TPJAC(1,1) = 0.0; TPJAC(1,2) = 0.0;
			TPJAC(1,3) = 0.0; TPJAC(1,4) = 0.0;
			TPJAC(1,5) = 0.0;
			TPJAC(1,6) = 1.0;
		}
	}
}
/*---------------------------------------------*/
void delayArgumentsFunc_02(double t, int m, double* y,
                                   int l, double* p,
                                   int nu, int omega, double* alpha,
                                   ddemUserProvidedJac* userProvidedJac,
																	 void* comm)
{
	// alphaIndices = 0, 1, 2, ..., (nu + omega - 1) 
	// yIndices = 0, 1, 2, ..., (nVariables - 1) 
	// pIndices = 0, 1, 2, ..., (nParameters - 1) 
	//*******************************************
	// set alpha[alphaIndices]
  alpha[0] = t - p[0];
  alpha[1] = t - p[0];
	if(userProvidedJac->needed)
	{
		//userProvidedJac->tIndependent = DDEM_TRUE;
		// or set TJAC(alphaIndices)
		// TJAC(0) = ;
		TJAC(0) = 1.0;
		TJAC(1) = 1.0;

		userProvidedJac->yIndependent = DDEM_TRUE;
		// or set YJAC(alphaIndices, yIndices)
		// YJAC(0,0) = ;

		//userProvidedJac->pIndependent = DDEM_TRUE;
		// or set PJAC(alphaIndices, pIndices)
		// PJAC(0,0) = ;
		PJAC(0,0) = - 1.0;
		PJAC(0,1) = 0.0; PJAC(0,2) = 0.0; PJAC(0,3) = 0.0;
		PJAC(0,4) = 0.0; PJAC(0,5) = 0.0; PJAC(0,6) = 0.0;
		PJAC(1,0) = - 1.0;
		PJAC(1,1) = 0.0; PJAC(1,2) = 0.0; PJAC(1,3) = 0.0;
		PJAC(1,4) = 0.0; PJAC(1,5) = 0.0; PJAC(1,6) = 0.0;
	}
}
/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
void getSystem_02(ddemSystem* system)
{
	system->create(2, 7, 1, 0);

	system->setF(f_02);

	system->setHistory(historyFunc_02);

	system->setT0(0.0);

	system->setDelayArguments(1, 1, delayArgumentsFunc_02, DDEM_FALSE);
}
/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
void getSimulationA_02(double TOL, ddemSimulationAttributes* simA)
{
	simA->parameters[0] = 21.0 / 50.0; // tau
	simA->parameters[1] = 29.0 / 10.0; // rho
	simA->parameters[2] = 1.0 / 10.0; // alpha
	simA->parameters[3] = 33.0 / 100.0; // phi_1 constant
	simA->parameters[4] = - 1.0 / 10.0 ; // phi_1 slope
	simA->parameters[5] = 111.0 / 50.0; // phi_2 constant
	simA->parameters[6] = 1.0 / 10.0; // phi_2 slope

	simA->tEnd = 30.0;

	simA->relTolerance = TOL;
	simA->absTolerance[0] = TOL;
	simA->absTolerance[1] = TOL;
}
