/* DOUBLE CHECKED = YES*/
#include <iostream>
using namespace std;
#include <string>
//#include <fstream>


#include "ddem_constraints.h"
#include "ddem_error_codes.h"
#include "ddem_trace.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
int ddemConstraints::create(int nParameters,
                int nLinearConstraints, int nNonlinearConstraints,
                double infinityBound)
{
	if(created_ == DDEM_FALSE){
		nParameters_ = nParameters;
		nLinearConstraints_ = nLinearConstraints;
		nNonlinearConstraints_ = nNonlinearConstraints;
		infinityBound_ = infinityBound;
		noLowerBound_ = - (infinityBound_ * 10);
		noUpperBound_= infinityBound_ * 10;

		NonlinearConstraints_ = NULL;
		nConstraints_ = nParameters_ + nLinearConstraints_ +
											 nNonlinearConstraints_;
		lowerBounds_ = new double[nConstraints_];
		upperBounds_ = new double[nConstraints_];
		for (int i = 0; i < nConstraints_; i++){
			lowerBounds_[i] = noLowerBound_;
			upperBounds_[i] = noUpperBound_;
		}
		if (nLinearConstraints_ > 0)
			linearConstraintsMatrix_ = new double[nLinearConstraints_ * nParameters_];
		else
			linearConstraintsMatrix_ = NULL;

		isFree_ = new int[nParameters_];
		for(int j = 0; j < nParameters_; j++)
			isFree_[j] = DDEM_TRUE;
		
		created_ = DDEM_TRUE;
		return DDEM_SUCCESS;
	}else{
		cout << "ddemConstraints::create can't be called more than once"
				<<"\n";
		return DDEM_FAILURE;
	}
}
/************************************************************/
ddemConstraints::ddemConstraints()
{
	lowerBounds_ = NULL;
	upperBounds_ = NULL;
	linearConstraintsMatrix_ = NULL;
	isFree_ = NULL;
	created_ = DDEM_FALSE;
}
/************************************************************/
ddemConstraints::~ddemConstraints()
{
	if(lowerBounds_ != NULL){delete lowerBounds_; lowerBounds_ = NULL;}
	if(upperBounds_ != NULL){delete upperBounds_; upperBounds_ = NULL;}
	if(linearConstraintsMatrix_ != NULL){delete linearConstraintsMatrix_; linearConstraintsMatrix_ = NULL;}
	if(isFree_ != NULL){delete isFree_; isFree_ = NULL;}
}
/************************************************************/
int ddemConstraints::isFree(int parameterIndex)
{
	return isFree_[parameterIndex];
}
/************************************************************/
void ddemConstraints::setSimpleLowerBound(int parameterIndex, double lowerBound)
{
  lowerBounds_[parameterIndex] = lowerBound;
}
/************************************************************/
void ddemConstraints::setSimpleUpperBound(int parameterIndex, double upperBound)
{
  upperBounds_ [parameterIndex] = upperBound;
}
/************************************************************/
void ddemConstraints::setSimpleEquality(int parameterIndex, double parameterFixedValue)
{
  lowerBounds_[parameterIndex] = parameterFixedValue;
  upperBounds_ [parameterIndex] = parameterFixedValue;
	isFree_[parameterIndex] = DDEM_FALSE;
}
/************************************************************/
void ddemConstraints::setLinearConstraintsMatrix(double linearConstraintsMatrix[])
{
  for (int j = 0; j < nLinearConstraints_ * nParameters_; j++)
		linearConstraintsMatrix_[j] = linearConstraintsMatrix[j];
}
/************************************************************/
void ddemConstraints::setLinearConstraintLowerBound(int linearConstraintIndex,
                                   double lowerBound)
{
  int baseIndex = nParameters_;
  lowerBounds_[baseIndex + linearConstraintIndex - 1] = lowerBound;
}
/************************************************************/
void ddemConstraints::setLinearConstraintUpperBound(int linearConstraintIndex,
                                   double upperBound)
{
  int baseIndex = nParameters_;
  upperBounds_[baseIndex + linearConstraintIndex - 1] = upperBound;
}
/************************************************************/
void ddemConstraints::setLinearConstraintEquality(int linearConstraintIndex,
                                 double equalityRightHandSide)
{
  int baseIndex = nParameters_;
  lowerBounds_[baseIndex + linearConstraintIndex - 1] = equalityRightHandSide;
  upperBounds_[baseIndex + linearConstraintIndex - 1] = equalityRightHandSide;
}
/************************************************************/
void ddemConstraints::setNonlinearConstraints(ddemNonlinearConstraints
                                   nonlinearConstraints)
{
  NonlinearConstraints_ = nonlinearConstraints;
}
/************************************************************/
void ddemConstraints::setNonlinearConstraintLowerBound(int nonlinearConstraintIndex,
                                      double lowerBound)
{
  int baseIndex = nParameters_ + nLinearConstraints_;
  lowerBounds_[baseIndex + nonlinearConstraintIndex - 1] = lowerBound;
}
/************************************************************/
void ddemConstraints::setNonlinearConstraintUpperBound(int nonlinearConstraintIndex,
                                      double upperBound)
{
  int baseIndex = nParameters_ + nLinearConstraints_;
  upperBounds_[baseIndex + nonlinearConstraintIndex - 1] = upperBound;
}
/************************************************************/
void ddemConstraints::setNonlinearConstraintEquality(int nonlinearConstraintIndex,
                                    double equalityRightHandSide)
{
  int baseIndex = nParameters_ + nLinearConstraints_;
  lowerBounds_[baseIndex + nonlinearConstraintIndex - 1]= equalityRightHandSide;
  upperBounds_[baseIndex + nonlinearConstraintIndex - 1]= equalityRightHandSide;
}
/************************************************************/
void ddemConstraints::computeNonlinearConstraints(int l, int needc[], double p[],
											                   double conf[], double cjac[],
																				 int flag)
{
    if (NonlinearConstraints_ != NULL)
			NonlinearConstraints_(l, nNonlinearConstraints_, needc, p,
		                     conf, cjac, flag);
}
