/* DOUBLE CHECKED = YES*/
#include <iostream>
using namespace std;
#include "ddem_fsens_system.h"
#include "ddem_error_codes.h"
#include "ddem_trace.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
ddemFSensSystem::ddemFSensSystem(ddemSystem* baseSystem, int nS, int* sIndices)
{
	baseSystem_ = baseSystem;
	nS_ = nS;
	sIndices_ = sIndices;
	createUserProvidedJacs_();
}
/************************************************************/
ddemFSensSystem::~ddemFSensSystem()
{
	destroyUserProvidedJacs_();
	sIndices_ = NULL;
	baseSystem_ = NULL;
}
/************************************************************/
int ddemFSensSystem::getNVariables()/*IMPLEMENTED*/
{
	if(baseSystem_->getOmega() == 0)/* retarded System */
		return baseSystem_->getNVariables() * (1 + nS_);
	else/* neutral System needs equations for y' because of y''*/
		return baseSystem_->getNVariables() * (1 + nS_ + 1);
}
/************************************************************/
int ddemFSensSystem::getNParameters()/*UNCHANGED*/
{
	return baseSystem_->getNParameters();
}
/************************************************************/
int ddemFSensSystem::getNHistorySegments()/*UNCHANGED*/
{
	return baseSystem_->getNHistorySegments();
}
/************************************************************/
int ddemFSensSystem::getNu()/*UNCHANGED*/
{
	return baseSystem_->getNu();
}
/************************************************************/
int ddemFSensSystem::getOmega()/*IMPLEMENTED*/
{
	if(baseSystem_->getOmega() == 0)/* retarded System */
		return baseSystem_->getNu();
	else
		return baseSystem_->getOmega();
}
/************************************************************/
int ddemFSensSystem::evalF(double t, double* y,
								 double* z,
								 double* p,
								 double* f,
								 void* comm)
{
	int nu = baseSystem_->getNu();
	int omega = baseSystem_->getOmega();
	int m = baseSystem_->getNVariables();
	int l = baseSystem_->getNParameters();

	double* zBaseSystem;
	zBaseSystem = new double[(nu + omega) * m];
	for(int i=0; i < nu + omega; i++)
		for(int j=0; j < m; j++)
			if(baseSystem_->getOmega() == 0)/* retarded */
				zBaseSystem[i * m + j] = z[i * (m * (1 + nS_)) + j];
			else/* neutral */
				zBaseSystem[i * m + j] = z[i * (m * (1 + nS_ + 1)) + j];
	userProvidedJac4F_->reset();
	userProvidedJac4F_->needed = DDEM_TRUE;
	userProvidedJac4F_->needed2nd = DDEM_FALSE;
	int retCode =
		baseSystem_->evalFBase(t, y,
											zBaseSystem,
											p,
											f,
											userProvidedJac4F_,
											comm);
	delete zBaseSystem; zBaseSystem = NULL;
	if(retCode!= DDEM_SUCCESS)
		return retCode;
	
	
	double* alphaBaseSystem;
	alphaBaseSystem = new double[nu + omega];
	userProvidedJac4DelayArguments_->reset();
	userProvidedJac4DelayArguments_->needed = DDEM_TRUE;
	userProvidedJac4DelayArguments_->needed2nd = DDEM_FALSE;
	retCode = baseSystem_->evalDelayArgumentsBase(t, y, p,
																			alphaBaseSystem,
																			userProvidedJac4DelayArguments_,
																			comm);
	delete alphaBaseSystem; alphaBaseSystem = NULL;
	if(retCode!= DDEM_SUCCESS)
		return retCode;
	for(int i=0; i < nS_; i++){
		double* s_i = &y[(i + 1) * m];
		double* f_i = &f[(1 + i) * m];
		for(int j=0; j < m; j++)/* f_i = 0 */
			f_i[j] = 0.0;
		for(int j=0; j < m; j++){/* f_i += par(f)/par(y) * s_i */
			double* par_Fj_par_Y = &userProvidedJac4F_->yJac[j * m];/*j'th row*/
			for(int col=0; col < m; col++)
				f_i[j] += par_Fj_par_Y[col] * s_i[col];
		}
		for(int k=0; k < nu + omega; k++){/* sum k = 1,..nu + omega */
			double* par_Alpha_k_par_Y;
			par_Alpha_k_par_Y = &userProvidedJac4DelayArguments_->yJac[k * m];
			double* par_Alpha_k_par_P;
			par_Alpha_k_par_P = &userProvidedJac4DelayArguments_->pJac[k * l];
			double acc = 0;
			for(int col=0; col < m; col++)/* acc += par(alpha_k)/par(Y)*s_i */
				acc += par_Alpha_k_par_Y[col] * s_i[col];
			acc += par_Alpha_k_par_P[sIndices_[i]];/* acc += ...*/
			double* yp_Alpha_k;
			int colCount;
			if(baseSystem_->getOmega() == 0){/* retarded */
				colCount = (1 + nS_) * m;
				yp_Alpha_k = &z[colCount * (nu + k)];
			}else{/* neutral */
				colCount = (1 + nS_ + 1) * m;
				yp_Alpha_k = &z[colCount * k + (1 + nS_) * m];
			}
			double*	s_i_Alpha_k =  &z[colCount * k +
																(1 + i) * m];
			double* accVec;
			accVec = new double[m];
			for(int q=0; q < m; q++)
				accVec[q] = yp_Alpha_k[q] * acc + s_i_Alpha_k[q];

			for(int j=0; j < m; j++){
				double* par_f_j_par_y_Alpha_k = &userProvidedJac4F_->zJac[j * (m * (nu + omega))
																																	+ k * m];
				for(int col=0; col < m ;col++)
					f_i[j]+= par_f_j_par_y_Alpha_k[col] * accVec[col];
			}
			delete accVec;
		}
		for(int j=0; j < m; j++)/* f_i += par(f)/par(p_i)*/
			f_i[j]+= userProvidedJac4F_->pJac[j * l + sIndices_[i]];
	}
	if(baseSystem_->getOmega() > 0){/*NEUTRAL System*/
		double* f_x = &f[(1 + nS_) * m];
		double* x = &y[(1 + nS_) * m];
		for(int i=0; i < m; i++){/* compute equations for x' = ... (x=y') */
			f_x[i] = 0.0;
			f_x[i] += userProvidedJac4F_->tJac[i];
			double* par_Fi_par_Y = &userProvidedJac4F_->yJac[i * m];/*i'th row*/
			for(int col=0; col < m; col++)
				f_x[i]+= par_Fi_par_Y[col] * x[col];
			for(int k=0; k < nu + omega; k++){/* sum k = 1,..nu + omega */
				double* par_Alpha_k_par_Y;
				par_Alpha_k_par_Y = &userProvidedJac4DelayArguments_->yJac[k * m];
				double acc = 0.0;
				for(int col=0; col < m; col++)/* par(alpha_k)/par(Y)*x */
					acc+= par_Alpha_k_par_Y[col] * x[col];
				acc+= userProvidedJac4DelayArguments_->tJac[k];
				double acc2 = 0.0;
				double* par_f_i_par_y_alpha_k = & userProvidedJac4F_->zJac[ i * (m * (nu + omega))
																																		+ k * m];
				int colCount = (1 + nS_ + 1) * m;
				double* yp_Alpha_k = &z[colCount * k + (1 + nS_) * m];
				for(int j=0; j < m; j++){
					acc2 += par_f_i_par_y_alpha_k[j] * yp_Alpha_k[j];
				}
				f_x[i] += acc2 * acc;
			}
		}
	}
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::evalPartialsDelayArgumentsBaseSystem(double t, double* y,
																double* p,
																int selIndex,
																double& parAlphaSelParT,
																double* parAlphaSelParY,
																double* parAlphaSelParP,
																void* comm)
{
	int nu = baseSystem_->getNu();
	int omega = baseSystem_->getOmega();
	int m = baseSystem_->getNVariables();
	int l = getNParameters(); 
	double* alphaBaseSystem;
	alphaBaseSystem = new double[nu + omega];
	userProvidedJac4DelayArguments_->reset();
	userProvidedJac4DelayArguments_->needed = DDEM_TRUE;
	userProvidedJac4DelayArguments_->needed2nd = DDEM_FALSE;
	int retCodeTemp =
	baseSystem_->evalDelayArgumentsBase(t, y, p,
																			alphaBaseSystem,
																			userProvidedJac4DelayArguments_,
																			comm);
	delete alphaBaseSystem; alphaBaseSystem = NULL;
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	
	parAlphaSelParT = userProvidedJac4DelayArguments_->tJac[selIndex]; 
	for(int i=0; i < m; i++)
		parAlphaSelParY[i] = userProvidedJac4DelayArguments_->yJac[selIndex * m + i];
	for(int i=0; i < nS_; i++)
		parAlphaSelParP[i] = userProvidedJac4DelayArguments_->pJac[selIndex * l + sIndices_[i]];
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::getNSelectedParameters()
{
	return nS_;
}
/************************************************************/
int ddemFSensSystem::evalDelayArguments(double t, double* y,
																double* p,
																double* alpha,
																void* comm)/*IMPLEMENTED*/
{
	int retCode = 
		baseSystem_->evalDelayArguments(t, y,
																p,
																alpha,
																comm);
	if(retCode!= DDEM_SUCCESS)
		return retCode;
	int nu = baseSystem_->getNu();
	int omega = baseSystem_->getOmega();
	if(baseSystem_->getOmega() ==0)
		for(int i=0; i < nu; i++)/* Artificially added delays*/
			alpha[nu + i] = alpha[i];
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::evalHistoryStart(int historySegmentIndex,
															double*p,
															double& historyStart,
															void* comm)/*UNCHANGED*/
{
	return baseSystem_->evalHistoryStart(historySegmentIndex,
																	p,
																	historyStart,
																	comm);
}
/************************************************************/
int ddemFSensSystem::evalPartialsHistoryStart(int historySegmentIndex,
															double*p,
															double* partialsHistoryStart,
															void* comm)
{
	userProvidedJac4HistoryStart_->reset();
	userProvidedJac4HistoryStart_->needed = DDEM_TRUE;
	userProvidedJac4HistoryStart_->needed2nd = DDEM_FALSE;
	double historyStart;
	int retCodeTemp =
	baseSystem_->evalHistoryStartBase(historySegmentIndex,
											p,
											historyStart,
											userProvidedJac4HistoryStart_,
											comm);
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	for(int i=0; i < nS_; i++)
		partialsHistoryStart[i] = userProvidedJac4HistoryStart_->pJac[sIndices_[i]];
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::getHistoryStartDiscDegree(int historySegmentIndex,
																	int& historyStartDiscDegree,
																	void* comm)
{	
	int retCodeTemp =
		baseSystem_->getHistoryStartDiscDegree(historySegmentIndex,
																	historyStartDiscDegree,
																	comm);
	if(retCodeTemp!= DDEM_SUCCESS)
		return retCodeTemp;
	
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::evalDerivHistory(int historySegmentIndex,
																double t,
                                double* p,
                                double* h,
															  void* comm)/*IMPLEMENTED*/
{
	userProvidedJac4History_->reset();
	userProvidedJac4History_->needed = DDEM_TRUE;
	if(baseSystem_->getOmega() > 0)/* NEUTRAL */
		userProvidedJac4History_->needed2nd = DDEM_TRUE;
	else
		userProvidedJac4History_->needed2nd = DDEM_FALSE;
	baseSystem_->evalHistoryBase(historySegmentIndex,
												 t,
                         p,
                         h,
                         userProvidedJac4History_,
												 comm);
	int m = baseSystem_->getNVariables();
	for(int i=0; i < m; i++)
		h[i] = userProvidedJac4History_->tJac[i];
	if(baseSystem_->getOmega() > 0){/* NEUTRAL */
		pJac2SensOfVar_(userProvidedJac4History_->tpJac, h);
		tJac2XOfVar_(userProvidedJac4History_->ttJac, h);
	}
	else
		for(int i=0; i < nS_; i++)
			for( int j=0; j < m; j++)
				h[(i + 1) * m + j] = 0.0; /* There is no need to compute them.*/
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::evalHistory(int historySegmentIndex,
																double t,
                                double* p,
                                double* h,
															  void* comm)/*IMPLEMENTED*/
{
	userProvidedJac4History_->reset();
	userProvidedJac4History_->needed = DDEM_TRUE;
	userProvidedJac4History_->needed2nd = DDEM_FALSE;
	int retCode =
		baseSystem_->evalHistoryBase(historySegmentIndex,
																	t,
																	p,
																	h,
																	userProvidedJac4History_,
																	comm);
	if(retCode!= DDEM_SUCCESS)
		return retCode;
	pJac2SensOfVar_(userProvidedJac4History_->pJac,
								h);
	if(baseSystem_->getOmega() > 0)/*NEUTRAL System*/
		tJac2XOfVar_(userProvidedJac4History_->tJac,
									h);
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::compT0(double* p,
           double& t0,
					void* comm)/*UNCHANGED*/
{
	return baseSystem_->compT0(p,
										t0,
										comm);
}
/************************************************************/
int ddemFSensSystem::compPartialsT0(double* p,
           double* partialsT0,
					void* comm)
{
	userProvidedJac4T0_->reset();
	userProvidedJac4T0_->needed = DDEM_TRUE;
	userProvidedJac4T0_->needed2nd = DDEM_FALSE;
	double t0;
	int retCodeTemp =
	baseSystem_->compT0Base(p,
           t0,
           userProvidedJac4T0_,
					comm);
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	for(int i=0; i < nS_; i++)
		partialsT0[i] = userProvidedJac4T0_->pJac[sIndices_[i]];
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::compT0DiscDegree(double* p,
               int& t0DiscDegree,
							void* comm)
{
	int retCodeTemp = baseSystem_->compT0DiscDegree(p,
														t0DiscDegree,
														comm);
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::compY0(double* p,
           double* y0,
					 double* historyStarts,
					 void* comm)/*IMPLEMENTED*/
{
	userProvidedJac4Y0_->reset();
	userProvidedJac4Y0_->needed = DDEM_TRUE;
	userProvidedJac4Y0_->needed2nd = DDEM_FALSE;
	int retCode = 
		baseSystem_->compY0Base(p,
											y0,
											historyStarts,
											userProvidedJac4Y0_,
											comm);
	if(retCode!= DDEM_SUCCESS)
		return retCode;
	pJac2SensOfVar_(userProvidedJac4Y0_->pJac,
								y0);
	if(baseSystem_->getOmega() > 0){/*NEUTRAL System*/
		double t0;
		compT0(p, t0, comm);
		int m = baseSystem_->getNVariables();
		int nu = baseSystem_->getNu();
		int omega = baseSystem_->getOmega();
		double* z0BaseSystem;
		z0BaseSystem = new double[(nu + omega) * m];
		retCode = baseSystem_->compZ0(t0, y0,
											 p,
											 z0BaseSystem,
											 historyStarts,
											 comm);
		if(retCode!= DDEM_SUCCESS)
			return retCode;
		retCode = baseSystem_->evalF(t0, y0,
												z0BaseSystem,
												p,
												&y0[(1 + nS_) * m],
												comm);
		delete z0BaseSystem;
		if(retCode!= DDEM_SUCCESS)
			return retCode;
	}
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemFSensSystem::locateBetweenHistoryStarts(double t,
																						double* historyStarts,
																						int& betweenHisroryIndex)
{
	return baseSystem_->locateBetweenHistoryStarts(t, historyStarts,
																						betweenHisroryIndex);
}
/************************************************************/
void ddemFSensSystem::pJac2SensOfVar_(double* pJac, double* var)
{
	int l = baseSystem_->getNParameters();
	int m = baseSystem_->getNVariables();
	for(int i=0; i < m; i++)
		for(int j=0; j < nS_; j++)
			var[(j + 1) * m + i] = 
				pJac[i * l + sIndices_[j]];
}
/************************************************************/
void ddemFSensSystem::tJac2XOfVar_(double* tJac, double* var)
{
	int m = baseSystem_->getNVariables();
	for(int i=0; i < m; i++)
		var[(1 + nS_) * m + i] = tJac[i]; 
}
/************************************************************/
void ddemFSensSystem::createUserProvidedJacs_()
{
	int nu = baseSystem_->getNu();
	int omega = baseSystem_->getOmega() + baseSystem_->getNu();
	int m = baseSystem_->getNVariables();
	int l = baseSystem_->getNParameters();
	userProvidedJac4DelayArguments_ = new ddemUserProvidedJac;
	userProvidedJac4DelayArguments_->tJac = new double[nu + omega];
	userProvidedJac4DelayArguments_->pJac = new double[(nu + omega) * l];
	userProvidedJac4DelayArguments_->yJac = new double[(nu + omega) * m];

	userProvidedJac4F_= new ddemUserProvidedJac;
	userProvidedJac4F_->tJac = new double[m];
	userProvidedJac4F_->pJac = new double[m * l];
	userProvidedJac4F_->yJac = new double[m * m];
	userProvidedJac4F_->zJac = new double[m * (nu + omega) * m];

	userProvidedJac4HistoryStart_ = new ddemUserProvidedJac;
	userProvidedJac4HistoryStart_->pJac = new double[l];

	userProvidedJac4History_ = new ddemUserProvidedJac;
	userProvidedJac4History_->pJac = new double[m * l];
	userProvidedJac4History_->tJac = new double[m];
	if(baseSystem_->getOmega() > 0){
		userProvidedJac4History_->tpJac = new double[m * l];
		userProvidedJac4History_->ttJac = new double[m];
	}

	userProvidedJac4T0_ = new ddemUserProvidedJac;
	userProvidedJac4T0_->pJac = new double[l];

	userProvidedJac4Y0_ = new ddemUserProvidedJac;
	userProvidedJac4Y0_->pJac = new double[m * l];
}
void ddemFSensSystem::destroyUserProvidedJacs_()
{
	delete userProvidedJac4DelayArguments_;
	delete userProvidedJac4F_;
	delete userProvidedJac4HistoryStart_;
	delete userProvidedJac4History_;
	delete userProvidedJac4T0_;
	delete userProvidedJac4Y0_;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
