/* DOUBLE CHECKED = NO*/
#include <iostream>
using namespace std;


#include <nag.h>
#include <nag_errlist.h>
#include <Nag/nage04.h>
#include <iomanip>

#include <math.h>
#include "ddem_parameter_estimator.h"
#include "ddem_error_codes.h"
#include "ddem_trace.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
ddemParameterEstimationStatistics::ddemParameterEstimationStatistics()
{
	nReferencedParameterValues = 0;
	finalObjectiveValue = -1;
}

/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
int ddemParameterEstimator::
         EstimateParameters(ddemSimulator* simulator,
														ddemSystem* system,
                            ddemData* data,
                            ddemConstraints* constraints,
                            double parameters[],
														double relTolerance,
														double* absTolerance,
														double vanishingConst,
														ddemParameterEstimationStatistics* statistics,
														int peSelect,
														void* commWithUFuncs )
{
	if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED)
		cout << "CALL TO ddemParameterEstimator::EstimateParameters\n";
	int retCodeTemp;
  /*****************************************/
  ddemSensitivityAnalyzer* sensitivityAnalyzer;
  sensitivityAnalyzer = new ddemSensitivityAnalyzer();
  /*****************************************/
  Integer m = data->nYs_;
  Integer n = system->getNParameters();
  Integer nclin = constraints->nLinearConstraints_;
  Integer ncnlinUser = constraints->nNonlinearConstraints_;
  double* a = constraints->linearConstraintsMatrix_/*={1.5}*/;
  Integer tda;
  if ( nclin > 0)
    tda = n;
  else
    tda =0;
   /*****************************************/
  int MaxC012Constraints = 2 * data->nGammas_;
  double* bl ;
  bl = new double[constraints->nConstraints_ + MaxC012Constraints];
  double* bu ;
  bu = new double[constraints->nConstraints_ + MaxC012Constraints];
  for (int i = 0; i < constraints->nConstraints_; i++){
    bl[i] = constraints->lowerBounds_[i];
    bu[i] = constraints->upperBounds_[i];
  }

  //int nC012 = 0;
  int* C012;
  C012 = new int[MaxC012Constraints];
  int* dataPointsSegmentsIndices = NULL;
	int* dataPointsSegmentsIndicesPrev = NULL;
	int* dataPointsSegmentsIndicesNew = NULL;
	int* dataPointsSegmentsIndicesTest = NULL;

  dataPointsSegmentsIndices = new int[data->nGammas_];
  int* blC012GammaIndices = NULL;
	int* buC012GammaIndices = NULL;
  blC012GammaIndices = new int[MaxC012Constraints];
  buC012GammaIndices = new int[MaxC012Constraints];
  /*****************************************/
	/* special attention */
  //int nImportantLambdas;
	//int* importantLambdasIndices;

	/*****************************************/
  double objf;
  double* f;
  f = new double[m];
  double* fjac;
  Integer tdfjac = n;
  fjac = new double[m * tdfjac];
  /*****************************************/
  NagError fail;
	if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED)
		SET_FAIL(fail);
	else
		INIT_FAIL(fail);
  //fail.print = TRUE;
  Nag_E04_Opt nagOptions;
  e04xxc(&nagOptions);
  nagOptions.inf_bound = constraints->infinityBound_;
	if(peSelect == DDEM_PE_VERY_SIMPLE){
		nagOptions.obj_deriv = FALSE;
		//nagOptions.verify_grad = Nag_NoCheck;
	}else{
		nagOptions.obj_deriv = TRUE;
		//nagOptions.verify_grad = Nag_SimpleCheck;
	}
	nagOptions.verify_grad = Nag_NoCheck;
	//nagOptions.nonlin_feas_tol = 1e-4;
	if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED){
		nagOptions.print_level = Nag_Soln_Iter;
		nagOptions.list = TRUE;
		nagOptions.print_deriv = Nag_D_Full;
		nagOptions.print_gcheck = TRUE;
	}else{
		nagOptions.print_level =  Nag_NoPrint;
		nagOptions.list = FALSE;
		nagOptions.print_deriv = Nag_D_NoPrint;
		nagOptions.print_gcheck = FALSE;
	}
	//nagOptions.f_prec = 1e-5;
	//nagOptions.nonlin_feas_tol = 1e-5;
  /* other settings of nagOptions */
  /*****************************************/
  /**********************************************************************/
  /* (1)Set-up and run problem simulation with initial parameter values */
  /*    to get the initilal configuration */
	//ddemSensitivitySolution* sensitivitySolution;
	//sensitivitySolution = new ddemSensitivitySolution();
	double tEnd = data->gammas_[data->nGammas_ - 1] * (1.0 + 1e-2) ;//+3.0 ;
	double tLeft = tEnd - (tEnd - data->gammas_[data->nGammas_ - 1])/5;

	ddemHelperComm* helperComm;
  helperComm = new ddemHelperComm;
	
	helperComm->statistics = statistics;

	helperComm->relTolerance = relTolerance;
	helperComm->absTolerance = absTolerance;
	helperComm->vanishingConst = vanishingConst;

	helperComm->parameterEstimator = this;

	helperComm->first = DDEM_TRUE;
	helperComm->lastParameters = NULL;


	helperComm->tLeft = tLeft;
	helperComm->peSelect = peSelect;
	helperComm->tEnd = tEnd;
	helperComm->commWithUFuncs = commWithUFuncs;
	helperComm->simulator = simulator;
  helperComm->system = system;
  helperComm->data = data;
  helperComm->constraints = constraints;
  helperComm->sensitivityAnalyzer = sensitivityAnalyzer;
  helperComm->dataPointsSegmentsIndices = dataPointsSegmentsIndices;
  helperComm->C012 = C012;

	helperComm->nImportantLambdas = -1;
	helperComm->importantLambdasIndices = NULL;

	if((peSelect == DDEM_PE_NEW)||(peSelect == DDEM_PE_NEW_CE)){
		int* needSensitivity;
		needSensitivity = new int[n];
		for(int i=0; i < n; i++)
			needSensitivity[i] = DDEM_TRUE;
		retCodeTemp =
			provideSensitivities_(simulator, system, tEnd,
																parameters,
																relTolerance,
																absTolerance,
																vanishingConst,
																needSensitivity,
																//sensitivitySolution,
																//nImportantLambdas,
																//importantLambdasIndices,
																commWithUFuncs, helperComm);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;
		/* (2)Based on the results from (1), setup additional constraints     */
		/*    needed for C0&1&2_continuity of the objective function, */
		/*    and call them "{C0,C1,C2}_forced constraints"  */
		/*           */

		setDataPointsSegmentsIndices_(helperComm->lastSensitivitySolution,
																	helperComm->nImportantLambdas,
																	helperComm->importantLambdasIndices,
																	data->gammas_,
																	data->nGammas_,
																	dataPointsSegmentsIndices);
		dataPointsSegmentsIndicesPrev = new int[data->nGammas_];
		dataPointsSegmentsIndicesNew = new int[data->nGammas_];
		dataPointsSegmentsIndicesTest = new int[data->nGammas_];

		for(int j=0; j < data->nGammas_; j++)
			dataPointsSegmentsIndicesPrev[j] = dataPointsSegmentsIndices[j];

		retCodeTemp =
			setC012Constraints_(helperComm->nC012, C012, blC012GammaIndices, buC012GammaIndices,
																				 bl, bu, data, constraints,
																					dataPointsSegmentsIndices,
																					helperComm->nImportantLambdas);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;
		delete needSensitivity;
	}else{
		helperComm->nC012 = 0;
	}/* end if */
  /*  REPEAT                                                     */
  //int found = DDEM_FALSE;
  int iterC = 0;
  do {
		/* (3)Run NAG's least squares fitter with user defined constraints +  */
		/*    "{C0+C1+C2}_forced constraints"*/
    /* applyNagNonlinearLeastSquares_(); */
    Nag_Comm nagComm;
    nagComm.p = (void*) helperComm;
		if(DDEM_TRACE_PARAMETER_ESTIMATOR_CONFIG && ((peSelect == DDEM_PE_NEW)||(peSelect == DDEM_PE_NEW_CE)))
		{
			cout <<"calling NAG ... \n";
			cout << "configuration =                                                     "; 
			for(int i=0; i < helperComm->data->nGammas_ ;i++)
					cout << helperComm->dataPointsSegmentsIndices[i] << " ";
			cout << "\n";
		}
    e04unc(m, n, nclin, ncnlinUser + helperComm->nC012, a, tda, bl, bu, data->Ys_,
      subfunctionsCalculator ,
      constraintsCalculator ,
      parameters , &objf, f, (double*)fjac, tdfjac,
      &nagOptions,
      &nagComm, &fail);
		statistics->finalObjectiveValue = objf;
		if(DDEM_TRACE_PARAMETER_ESTIMATOR){
			cout <<"after return from NAG\n";
			//cout << "parameter values :" << endl;
			for(int i=0; i < system ->getNParameters() ;i++){
				cout << "p[" << i << "]= ";
				cout.setf(ios::fixed); cout << setprecision(16);
				cout << parameters[i] << "\n";
				//printf("p[%d]= %.16f\n", i, parameters[i]);
			}
			cout <<"objective value =             ";
			cout.setf(ios::scientific); cout << setprecision(16);
			cout << objf << "\n";
			//printf("objective value =             %e\n", objf);
			//cout << "crash test value  =                    " << 1.0 * 2.0 << "\n";
			//printf("objective value =%.20f\n", objf);
			//cout << "objective value =                    " << n <<"\n";
			//cout << "------------------------------------------------------------\n";
		}
		if(fail.code == NE_USER_STOP)
			return fail.errnum;
		if((peSelect == DDEM_PE_VERY_SIMPLE) || (peSelect == DDEM_PE_SIMPLE))
			if(fail.code == 0)
				return DDEM_SUCCESS;
			else
				return (DDEM_ERR_NAG_BASE - fail.code);
		//	break;
		/* (4)Check the activity of "{C0&C1&C2}_forced constraints" in the */
		/*    optimum found from (3) */
		if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED){
			cout << "CALL2 TO ddemParameterEstimator::EstimateParameters\n";
			cout << "States of the constraints\n";
			for(int i=0; i< helperComm->nC012; i++) 
				cout << nagOptions.state[constraints->nConstraints_ + i] <<"\n";
		}
		int* activeStates = NULL;
		int* activeIndices = NULL;
		activeStates = new int[helperComm->nC012];
		activeIndices = new int[helperComm->nC012];
		int nActives = 0;
		#define	DDEM_INACTIVE 0
		#define DDEM_ACTIVE_AT_LOWER_BOUND 1
		#define DDEM_ACTIVE_AT_UPPER_BOUND 2
    for (int i=0; i < helperComm->nC012 ; i++){
      if (nagOptions.state[constraints->nConstraints_ + i] == 1)/* ||
          (nagOptions.state[constraints->nConstraints_ + i] == 0 &&
           (nagOptions.conf[constraints->nNonlinearConstraints_ + i] -
           bl[constraints->nConstraints_ + i] <
           nagOptions.f_prec * (1 + fabs(bl[constraints->nConstraints_ + i]))
           )))*/{/* Active at */
      /* lower bound*/
        //dataPointsSegmentsIndices[blC012GammaIndices[i]]+=1;
				activeStates[i] = DDEM_ACTIVE_AT_LOWER_BOUND;
				activeIndices[nActives] = i;
				nActives += 1;
      }else if (nagOptions.state[constraints->nConstraints_ + i] == 2)/* ||
          (nagOptions.state[constraints->nConstraints_+i] == 0 &&
           (bu[constraints->nConstraints_ + i]
           - nagOptions.conf[constraints->nNonlinearConstraints_ + i] <
           nagOptions.f_prec * (1+fabs(bu[constraints->nConstraints_ + i]))
           )))*/{
      /*Active at upper bound*/
        //dataPointsSegmentsIndices[buC012GammaIndices[i]]-=1;
				activeStates[i] = DDEM_ACTIVE_AT_UPPER_BOUND;
				activeIndices[nActives] = i;
				nActives += 1;
      }else
				activeStates[i] = DDEM_INACTIVE;
    }
		if(nActives == 0){
			if(DDEM_TRACE_PARAMETER_ESTIMATOR)
			{
				cout <<"No active constraints were found." << endl;
				cout << "Local optimum was found" << endl; 
			}
			break;/*do/while*/
		}
		/*------------------------------------*/
		int* decreasesMin = NULL;
		decreasesMin = new int[nActives];
		int* activeGammaIndices = NULL;
		activeGammaIndices = new int[nActives];
		for(int j=0; j < data->nGammas_; j++)
			dataPointsSegmentsIndicesTest[j] = dataPointsSegmentsIndices[j];
    for (int k=0; k < nActives ; k++)
			if(activeStates[activeIndices[k]] == DDEM_ACTIVE_AT_LOWER_BOUND){
				activeGammaIndices[k] = blC012GammaIndices[activeIndices[k]];
				dataPointsSegmentsIndicesTest[activeGammaIndices[k]] += 1;
			}else{/*DDEM_ACTIVE_AT_UPPER_BOUND*/
				activeGammaIndices[k] = buC012GammaIndices[activeIndices[k]];	
				dataPointsSegmentsIndicesTest[activeGammaIndices[k]] -= 1;
			}
		/*----------------------*/
		double* fNew;
		fNew = new double[data->nYs_];
		double* fOld;
		fOld = f;

		int* needSensitivity;
		needSensitivity = new int[n];
		for(int i=0; i < n; i++)
			needSensitivity[i] = DDEM_TRUE;
		retCodeTemp =
			provideSensitivities_(simulator, system, tEnd,
                              parameters,
															relTolerance, absTolerance, vanishingConst,
															needSensitivity,
 															commWithUFuncs,
															helperComm);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;
		retCodeTemp =
			configPreservEvaluate_(helperComm->sensitivityAnalyzer,
																				 data->nGammas_,
                                         data->gammas_,
                                         data->nYs_,
                                         data->gammasIndices_,
                                         data->YsIndices_,
                                         dataPointsSegmentsIndicesTest,
                                         n,// of parameters
                                         parameters,
                                         fNew,
                                         fjac,
                                         2,// flag
																				 helperComm->lastSensitivitySolution,
																				 helperComm->nImportantLambdas,
																				 helperComm->importantLambdasIndices,
																				 simulator,
																				 needSensitivity,
																				 system,
																				 tEnd,
																				 tLeft);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;
		/*----------------------*/
		double* fSumNew;
		fSumNew = new double[nActives];
		double* fSumOld;
		fSumOld = new double[nActives];
		int nActiveDecrease = 0;
		for(int k=0; k < nActives; k++){
			if(helperComm->lastSensitivitySolution->lambdasDiscDegrees[
											             helperComm->importantLambdasIndices[C012[activeIndices[k]]]] == 0){
				fSumNew[k] = 0.0;
				fSumOld[k] = 0.0;
				for(int i=data->invertGammaIndicesFirst_[activeGammaIndices[k]];
					i <= data->invertGammaIndicesLast_[activeGammaIndices[k]]; i++){
					fSumNew[k] += 0.5 * (data->Ys_[i] - fNew[i]) * (data->Ys_[i] - fNew[i]);
					fSumOld[k] += 0.5 * (data->Ys_[i] - fOld[i]) * (data->Ys_[i] - fOld[i]);
				}
				if(fSumNew[k] <= fSumOld[k] * (1.0 + 1e-10))
				{
					decreasesMin[k] = DDEM_TRUE;
					nActiveDecrease += 1;
				}
				if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED)
					cout << "k = " << k << " fSumNew = " << fSumNew[k] << " fSumOld = " << fSumOld[k] <<"\n";
			}else{
					decreasesMin[k] = DDEM_TRUE;
					nActiveDecrease += 1;
			}
		}
		/*---------------------*/
		if(nActiveDecrease == 0){
			if(DDEM_TRACE_PARAMETER_ESTIMATOR)
			{
				cout <<"No usefull active constraints were found." << endl;
				cout << "<<<<<<<<<<<< Local optimum was found >>>>>>>>>>." << endl << endl; 
			}
			break;/*do/while*/
		}

		for(int j=0; j < data->nGammas_; j++)
			dataPointsSegmentsIndicesNew[j] = dataPointsSegmentsIndices[j];
    for (int k=0; k < nActives ; k++)
			if(decreasesMin[k] == DDEM_TRUE)
				if(activeStates[activeIndices[k]] == DDEM_ACTIVE_AT_LOWER_BOUND)
					dataPointsSegmentsIndicesNew[blC012GammaIndices[activeIndices[k]]] += 1;
				else/*DDEM_ACTIVE_AT_UPPER_BOUND*/
					dataPointsSegmentsIndicesNew[buC012GammaIndices[activeIndices[k]]] -= 1;
		/*------------------------------------*/
		int isLooping = DDEM_TRUE;
		for(int i=0; i < data->nGammas_; i++)
			if(dataPointsSegmentsIndicesNew[i] != dataPointsSegmentsIndicesPrev[i]){
				isLooping = DDEM_FALSE;
				break;/* for*/
			}
		if(isLooping == DDEM_TRUE){
			if(DDEM_TRACE_PARAMETER_ESTIMATOR)
			{
				cout <<"Returning to previouse configuration... \n";
				cout << "configuration =                                                     "; 
				for(int i=0; i < helperComm->data->nYs_ ;i++)
						cout << dataPointsSegmentsIndicesPrev[i]<<" ";
				cout << "\n";
				cout << "<<<<<<<<<<<< Local optimum was found >>>>>>>>>>." << endl << endl; 
			}
 			break;/*do/while*/
		}
		for(int i=0; i < data->nGammas_; i++){
			dataPointsSegmentsIndicesPrev[i] = dataPointsSegmentsIndices[i];
			dataPointsSegmentsIndices[i] = dataPointsSegmentsIndicesNew[i];
		}
		/*-------------------------------------*/
		retCodeTemp =
			setC012Constraints_(helperComm->nC012, C012, blC012GammaIndices, buC012GammaIndices,
                                       bl, bu, data, constraints,
                                        dataPointsSegmentsIndices,
                                        helperComm->nImportantLambdas);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;
		/* (5)If there is not any, then the local optimum was found -> FOUND */
		/*   ((the objective function is not increased at the passage of active */
		/*   constraints and therefore, the new optimum is at least as good */
		/*   as the one that we had before passing the constraints)) */
		/* (6)Looking at the set of visited configurations, decide whether an */
		/*    optimum has been found or we should proceed*/
		/*    case  I: optimum -> FOUND  */
		/*    case II: based on the active "C0&C1&2_forced constraints",
		/*             find a new set of "C0&C1&2_forced constraints" which */
		/*             define  a "new configuration"  ->  continue with (3) */
		/* UNTIL a local optimum is found */
    iterC += 1;
		if (iterC >= 10){
			return DDEM_ERR_PARAMEST_MAX_ITER_REACHED;
		}
		delete fSumOld;
		delete fSumNew;
		delete needSensitivity;
		delete fNew;
		if(decreasesMin != NULL) delete decreasesMin;
		if(activeGammaIndices != NULL) delete activeGammaIndices;
		if(activeStates != NULL) delete activeStates;
		if(activeIndices != NULL) delete activeIndices;
  } while (DDEM_TRUE);
	

	
	if(dataPointsSegmentsIndicesTest != NULL) delete dataPointsSegmentsIndicesTest;
	if(dataPointsSegmentsIndicesPrev != NULL)	delete dataPointsSegmentsIndicesPrev;
	if(dataPointsSegmentsIndicesNew != NULL)	delete dataPointsSegmentsIndicesNew;
	if(helperComm->importantLambdasIndices != NULL)	delete helperComm->importantLambdasIndices;
	if(helperComm->lastSensitivitySolution != NULL)	delete helperComm->lastSensitivitySolution;
	if(helperComm->lastParameters != NULL)	delete helperComm->lastParameters;

	delete helperComm;

  delete C012;
  delete dataPointsSegmentsIndices;
  delete blC012GammaIndices;
  delete buC012GammaIndices;
  delete f;
  delete fjac;

	delete bl;
	delete bu;
	delete sensitivityAnalyzer;
	return DDEM_SUCCESS;
}
/************************************************************/
/*CHECKED = ALMOST*/
int ddemParameterEstimator::setC012Constraints_(int& nC012, int* C012,
                                        int* blC012GammaIndices,
                                        int* buC012GammaIndices,
                                        double* bl, double * bu,
                                        ddemData* data,
                                        ddemConstraints* constraints,
                                        int* dataPointsSegmentsIndices,
                                        int nImportantLambdasActual)
{
	int nImportantLambdas;
	if (dataPointsSegmentsIndices[data->nGammas_ - 1] == (nImportantLambdasActual -1 ))
		nImportantLambdas = nImportantLambdasActual + 1;
	else
		nImportantLambdas = nImportantLambdasActual;

  #define INVALID_INDEX -1
  int prevLeftLambdaIndex = INVALID_INDEX;
  int prevRightLambdaIndex = INVALID_INDEX;
  int iC012 = -2 ;/*last left index that was set*/
  /*   o : data points */
  /*   + : discontinuity points */
  for (int j=0; j < data->nGammas_; j++){
    int leftLambdaIndex = dataPointsSegmentsIndices[j];
    int rightLambdaIndex = leftLambdaIndex + 1;
    if (leftLambdaIndex == prevLeftLambdaIndex){
    /*    + o o +    */
      bl[constraints->nConstraints_+iC012+1] = data->gammas_[j];
      blC012GammaIndices[iC012+1] = j;
      /*bu[constraints_->nConstraints_+iC012+1] = constraints_->noUpperBound_;*/
      iC012+=0;
    }else if (leftLambdaIndex == prevRightLambdaIndex){/*new rightLambdaIndex*/
    /*    + o + o +    */
      C012[iC012+2] = rightLambdaIndex;
      bu[constraints->nConstraints_+iC012+1] = data->gammas_[j];
      buC012GammaIndices[iC012+1] = j;
      
      bl[constraints->nConstraints_+iC012+2] = data->gammas_[j];
      blC012GammaIndices[iC012+2] = j;
      bu[constraints->nConstraints_+iC012+2] = constraints->noUpperBound_;
      buC012GammaIndices[iC012+2] = INVALID_INDEX;
      iC012+=1;
    }else{/* new  leftLambdaIndex and rightLambdaIndex*/
    /*    + o + + + + o +    */
      C012[iC012+2] = leftLambdaIndex;
      bl[constraints->nConstraints_+iC012+2] = constraints->noLowerBound_;
      blC012GammaIndices[iC012+2] = INVALID_INDEX;
      bu[constraints->nConstraints_+iC012+2] = data->gammas_[j];
      buC012GammaIndices[iC012+2] = j;
      C012[iC012+3] = rightLambdaIndex;
      bl[constraints->nConstraints_+iC012+3] = data->gammas_[j];
      blC012GammaIndices[iC012+3] = j;
      bu[constraints->nConstraints_+iC012+3] = constraints->noUpperBound_;
      buC012GammaIndices[iC012+3] = INVALID_INDEX;
      iC012+=2;
    }
    prevLeftLambdaIndex = leftLambdaIndex;
    prevRightLambdaIndex = rightLambdaIndex;
  }
  if (dataPointsSegmentsIndices[data->nGammas_ - 1] == (nImportantLambdas - 1))
    /* the last data point is located after the last important      */
    /* discontinuity point -> the last inequality is fictitious and */
    /* should be removed */
    /* + o + + + o o +(last) o o o(last)   +(fictitious discontinuity point) */
    /*                                  <=                                   */
    nC012 = iC012+1;
  else
    nC012 = iC012+2;
	return DDEM_SUCCESS;
}
/************************************************************/
/*CHECKED = YES*/
int ddemParameterEstimator::provideSensitivities_(ddemSimulator* simulator,
															ddemSystem* system, double tEnd,
                              double parameters[],
															double relTolerance,
															double* absTolerance,
															double vanishingConst,
															int needSensitivity[],
                              //ddemSensitivitySolution*& sensitivitySolution,
															//int& nImportantLambdas,
															//int*& importantLambdasIndices,
															void* commWithUFuncs,
															ddemHelperComm* helperComm)
{
	//cout << "enter (ddemParameterEstimator::provideSensitivities)"<<"\n";
	int equal = DDEM_TRUE;
	int l = system->getNParameters();
	if(helperComm->first == DDEM_TRUE){
		equal = DDEM_FALSE;
		helperComm->lastParameters = new double[l];
	}else{
		for(int i = 0; i < l; i++)
			if( parameters[i] != helperComm->lastParameters[i]){
				equal = DDEM_FALSE;
				break;
			}
	}
	if(equal == DDEM_FALSE){
		if(helperComm->first == DDEM_FALSE){
			delete helperComm->lastSensitivitySolution;
			helperComm->lastSensitivitySolution = NULL;
			delete helperComm->importantLambdasIndices;
			helperComm->importantLambdasIndices = NULL;
		}
		helperComm->lastSensitivitySolution = new ddemSensitivitySolution();
		int retCodeTemp = 
			helperComm->sensitivityAnalyzer->computeSensitivities(simulator,
														system, tEnd,
                            parameters,
														relTolerance, absTolerance, vanishingConst,
														needSensitivity,                      
														commWithUFuncs,
														helperComm->lastSensitivitySolution);
		helperComm->statistics->nReferencedParameterValues += 1;
		helperComm->statistics->nFEvals += helperComm->lastSensitivitySolution->statistics->nFEvals;
		if(DDEM_TRACE_PARAMETER_ESTIMATOR){
			cout << "Function calculation requested for :" << endl; 
			for(int i=0; i < l ;i++){
				cout << "p[" << i << "] =                                     ";
				cout.setf(ios::fixed); cout << setprecision(16);
				cout << parameters[i] << "\n";
				//printf("p[%d] =                                     %.16f\n", i, parameters[i]);
			}
		}
		if(retCodeTemp != DDEM_SUCCESS){
			delete helperComm->lastSensitivitySolution;
			helperComm->lastSensitivitySolution = NULL;
			return retCodeTemp;
		}
		setImportantLambdasIndices_(helperComm->lastSensitivitySolution,
																helperComm->nImportantLambdas,
																helperComm->importantLambdasIndices,
																2/*importanceLevel*/);
		for(int i = 0; i < l; i++)
			helperComm->lastParameters[i] = parameters[i];
	}
	//helperComm->lastSensitivitySolution = helperComm->lastSensitivitySolution;
	helperComm->first = DDEM_FALSE;
	//cout << "leave (ddemParameterEstimator::provideSensitivities)"<<"\n";
	return DDEM_SUCCESS;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void subfunctionsCalculator(Integer m, Integer n, double p[],
    				                     double f[], double fjac[], Integer tdfjac,
		                             Nag_Comm *comm)
{
	if(DDEM_TRACE_PARAMETER_ESTIMATOR){
		cout << "CALL TO subfunctionsCalculator\n";// << p[0]<<"\n";
		for(int i=0; i < n; i++){
			cout << "p[" << i << "]= ";
			cout.setf(ios::fixed); cout << setprecision(16);
			cout << p[i] << "\n";
			//printf("p[%d] = %.16f\n", i,  p[i]);
		}
	}
 /*****************************************/
  /*Run Sensitivity Analyzer with input p*/
  /*to get values used in the computation of f, fjac*/
  ddemHelperComm* hC;
  hC = (ddemHelperComm*)comm->p;
	int retCodeTemp;
  /*****************************************/
	if(hC->peSelect == DDEM_PE_VERY_SIMPLE){
		ddemSimulationSolution* simulationSolution;
		simulationSolution = new ddemSimulationSolution();
		retCodeTemp = 
			hC->simulator->simulate(hC->system, hC->tEnd, p,
															hC->relTolerance, hC->absTolerance,
															hC->vanishingConst,
															hC->commWithUFuncs,
															simulationSolution);
		hC->statistics->nReferencedParameterValues += 1;
		hC->statistics->nFEvals += simulationSolution->statistics->nFEvals;
		if(retCodeTemp != DDEM_SUCCESS)
		{
			comm->flag = retCodeTemp;
			delete simulationSolution;
			return;
		}
		double* y = new double[simulationSolution->getNVariables()];
		for(int i=0; i < hC->data->nYs_; i++){
			int gammaIndex = hC->data->gammasIndices_[i];
			if(gammaIndex < 0 || gammaIndex >= hC->data->nGammas_)
				cout << "subfunctionsCalculator :: Unexpected Situation! \n";
		  double t = hC->data->gammas_[gammaIndex];
			retCodeTemp =
				hC->simulator->evaluateSolution(t, y, simulationSolution, DDEM_INTERP, -1);
			if(retCodeTemp != DDEM_SUCCESS)
			{
				comm->flag = retCodeTemp;
				delete simulationSolution;
				delete y;
				return;
			}
			int yIndex =  hC->data->YsIndices_[i];
			f[i] = y[yIndex];
		}
		delete y;
		delete simulationSolution;
		return;
	}
	/*else */
  /*****************************************/
	//ddemSensitivitySolution* sensitivitySolution;
	/* special attention*/
	//int nImportantLambdas;
	//int* importantLambdasIndices;
	//sensitivitySolution = new ddemSensitivitySolution();
  int* needSensitivity;
  needSensitivity = new int[n];
	for(int i=0; i < n; i++)
		needSensitivity[i] = DDEM_TRUE;
	//cout << " n = " << n << endl;
	retCodeTemp =
		hC->parameterEstimator->provideSensitivities_(hC->simulator,
															hC->system, hC->tEnd,
                              p,
															hC->relTolerance, hC->absTolerance,
															hC->vanishingConst,
															needSensitivity,
															hC->commWithUFuncs,
															hC);
	if(retCodeTemp != DDEM_SUCCESS)
	{
		comm->flag = retCodeTemp;
		delete needSensitivity;
		return;
	}
	//sensitivitySolution->setImportantLambdasIndices(2/*importanceLevel*/);
	if(hC->peSelect == DDEM_PE_NEW_CE){
		//cout << "HERE 1" << endl;
		retCodeTemp =
		    hC->parameterEstimator->configPreservEvaluate_(hC->sensitivityAnalyzer,
																						 hC->data->nGammas_,
                                             hC->data->gammas_,
                                             hC->data->nYs_,
                                             hC->data->gammasIndices_,
                                             hC->data->YsIndices_,
                                             hC->dataPointsSegmentsIndices,
                                             n,p,
                                             f, fjac, 2/*flag*/,
																						 hC->lastSensitivitySolution,
																						 hC->nImportantLambdas,
																						 hC->importantLambdasIndices,
																						 hC->simulator,
																						 needSensitivity,
																						 hC->system,
																						 hC->tEnd,
																						 hC->tLeft);
		//cout << "HERE 2" << endl;
		if(retCodeTemp != DDEM_SUCCESS)
		{
			//cout << "HERE 3" << endl;
			comm->flag = retCodeTemp;
			delete needSensitivity;
			return;
		}
	}else{
		retCodeTemp =
			hC->parameterEstimator->simpleEvaluate_(hC->sensitivityAnalyzer,
																						 hC->data->nGammas_,
                                             hC->data->gammas_,
                                             hC->data->nYs_,
                                             hC->data->gammasIndices_,
                                             hC->data->YsIndices_,
                                             /*hC->dataPointsSegmentsIndices,*/
                                             n,p,
                                             f, fjac, 2/*flag*/,
																						 hC->lastSensitivitySolution,
																						 hC->nImportantLambdas,
																						 hC->importantLambdasIndices,
																						 hC->simulator,
																						 needSensitivity,
																						 hC->system,
																						 hC->tEnd,
																						 hC->tLeft);
		if(retCodeTemp != DDEM_SUCCESS)
		{
			comm->flag = retCodeTemp;
			delete needSensitivity;
			return;
		}
	}
	delete needSensitivity;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*CHECKED = ALMOST*/
void constraintsCalculator(Integer n, Integer ncnlin, Integer needc[],
		                            double p[], double conf[], double cjac[],
		                            Nag_Comm *comm)
{
	if(DDEM_TRACE_PARAMETER_ESTIMATOR){
		cout << "CALL TO constraintsCalculator\n";//<<p[0]<<"\n" ;
		for(int i=0; i < n; i++){
				cout << "p[" << i << "]= ";
				cout.setf(ios::fixed); cout << setprecision(16);
				cout << p[i] << "\n";
			//printf("p[%d] = %.16f\n", i, p[i]);
		}
	}
/* 	cout << "CALL1.1 TO constraintsCalculator\n"<<needc[0]<<" "<<needc[1] <<"\n" ; */
	int retCodeTemp;
  /*****************************************/
  ddemHelperComm* hC;
  hC = (ddemHelperComm*)comm->p;
  int nNonlinearConstraints = hC->constraints->getnNonlinearConstraints();
  if (nNonlinearConstraints > 0)
	 hC->constraints->computeNonlinearConstraints((int)n, (int*)needc,
																						(double*) p, (double*)conf,
																						(double*) cjac,
																						(int)comm->flag);
	if((hC->peSelect == DDEM_PE_VERY_SIMPLE)|| (hC->peSelect == DDEM_PE_SIMPLE))
		return;
	/*else */
	//ddemSensitivitySolution* sensitivitySolution;
	/* special attention */
	//int nImportantLambdas;
  //int* importantLambdasIndices;
	//sensitivitySolution = new ddemSensitivitySolution();
  int* needSensitivity;
  needSensitivity = new int[n];
	for(int i=0; i < n; i++)
		needSensitivity[i] = DDEM_TRUE;
	
	retCodeTemp =
		hC->parameterEstimator->provideSensitivities_(hC->simulator,
															hC->system, hC->tEnd,
                              p,
															hC->relTolerance, hC->absTolerance,
															hC->vanishingConst,
															needSensitivity,
                              //sensitivitySolution,
															//nImportantLambdas,
															//importantLambdasIndices,
															hC->commWithUFuncs,
															hC);

	if(retCodeTemp != DDEM_SUCCESS)
	{
		comm->flag = retCodeTemp;
		delete needSensitivity;
		return;
	}
	//sensitivitySolution->setImportantLambdasIndices(2/*importanceLevel*/);
  for (int i=0; i < hC->nC012; i++){
		if(hC->C012[i] < hC->nImportantLambdas){/* lambda <= tEnd */
			double lambda = hC->lastSensitivitySolution->meshPoints[
                          hC->lastSensitivitySolution->lambdasIndices[
                          hC->importantLambdasIndices[
                          hC->C012[i]
                          ]]];
			//cout << "lambda = " << lambda <<"\n";
			if(lambda < hC->tLeft) /* lambda <= tLeft */
				conf[nNonlinearConstraints + i] = lambda;
			else{/* tLeft<= lambda <= tEnd */
				double theta_0_lambda = ddem_theta_0(hC->tEnd, hC->tLeft, lambda);
				conf[nNonlinearConstraints + i] = theta_0_lambda * lambda +
																				(1.0 - theta_0_lambda) * hC->tEnd;
				if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED){
					cout << "<<<<<< SPECIAL SITUATION in constraintsCalculator >>>>>>> \n";
					cout << "tLeft < lambda <= tEnd \n";
				}
			//cout <<"i = "<< i << "-----------------------------------------------------YES HAPPENED 111\n";
			}
		}else{/* tEnd <= lambda */
			conf[nNonlinearConstraints + i] = hC->tEnd;
			//cout <<"i = "<< i <<  "------------------------------------------------------YES HAPPENED 222\n";
			if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED){
				cout << "<<<<<< SPECIAL SITUATION in constraintsCalculator >>>>>>> \n";
				cout << "tEnd <= lambda \n";
			}
		}
  }
  /*****************************************/
	#define CJAC(I,J) cjac[(I)*n + (J)]
  for (int i=0; i < hC->nC012; i++){
		if(hC->C012[i] < hC->nImportantLambdas){/* lambda <= tEnd */
			double lambda = hC->lastSensitivitySolution->meshPoints[
                          hC->lastSensitivitySolution->lambdasIndices[
                          hC->importantLambdasIndices[
                          hC->C012[i]
                          ]]];
			if(lambda < hC->tLeft) /* lambda <= tLeft */
				for (int j=0; j < n; j++)
			    CJAC(nNonlinearConstraints + i,j) = hC->lastSensitivitySolution->
				                                lambdasJac[
					                              hC->importantLambdasIndices[hC->C012[i]]][j];
			else{/* tLeft<= lambda <= tEnd */
				double theta_0_lambda = ddem_theta_0(hC->tEnd, hC->tLeft, lambda);
				double theta_0_p_lambda = ddem_theta_0_p(hC->tEnd, hC->tLeft, lambda);
				for (int j=0; j < n; j++){
					double par_lambda_par_p_j = hC->lastSensitivitySolution->
			                                lambdasJac[
				                              hC->importantLambdasIndices[hC->C012[i]]][j];
			    CJAC(nNonlinearConstraints + i,j) = theta_0_p_lambda *
																						 par_lambda_par_p_j *
																						 (lambda - hC->tEnd) +
																						 theta_0_lambda * par_lambda_par_p_j;
				}
				//cout <<"i = "<< i <<  "--------------------------------------------YES HAPPENED JAC 111\n";
			}
		}else{/* tEnd <= lambda */
			for (int j=0; j < n; j++)
				CJAC(nNonlinearConstraints + i,j) = 0.0;
				//cout <<"i = "<< i <<  "--------------------------------------------YES HAPPENED JAC 222\n";
    }
  }
	delete needSensitivity;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/************************************************************/
/*CHECKED = YES*/
void ddemParameterEstimator::setImportantLambdasIndices_(ddemSensitivitySolution* sensitivitySolution,
																												 int& nImportantLambdas,
																												 int*& importantLambdasIndices,
																												 int importanceLevel)
{
  int count=0;
  for (int i = 0; i < sensitivitySolution->nLambdas; i++)
    if ( sensitivitySolution->lambdasDiscDegrees[i] <= importanceLevel)
      count+=1;
  nImportantLambdas = count;
	if(importantLambdasIndices != NULL){
		delete importantLambdasIndices;
		importantLambdasIndices = NULL;
	}
	if(nImportantLambdas > 0){
		importantLambdasIndices = new int[nImportantLambdas];
		count=0;
		for (int i = 0; i < sensitivitySolution->nLambdas; i++)
			if (sensitivitySolution->lambdasDiscDegrees[i] <= importanceLevel){
				importantLambdasIndices[count] = i;
				count+=1;
			}
	}
}
/************************************************************/
/*CHECKED = YES*/
int ddemParameterEstimator::setDataPointsSegmentsIndices_(ddemSensitivitySolution* sensitivitySolution,
																													int nImportantLambdas,
																												  int* importantLambdasIndices,
																													double* gammas, int nGammas,
																												  int* dataPointsSegmentsIndices)
{
  int prevFoundIndex = 0;
  int error = DDEM_FALSE;
  for (int i = 0; i < nGammas; i++){
    int found = DDEM_FALSE;
    int candidateIndex = prevFoundIndex;
		if(candidateIndex >= nImportantLambdas){
		  cout << "(ddemParameterEstimator::setDataPointsSegmentsIndices)\n" <<
										" unexpected situation! \n";
			error = DDEM_TRUE;
			break;
		}
    while (!found && !error){
      double candidateLeft =  sensitivitySolution->meshPoints[
                     sensitivitySolution->lambdasIndices[
                    importantLambdasIndices[
                    candidateIndex
                    ]]];
      double candidateRight;
      if(candidateIndex + 1 < nImportantLambdas)
        candidateRight = sensitivitySolution->meshPoints[
                    sensitivitySolution->lambdasIndices[
                    importantLambdasIndices[
                    candidateIndex + 1
                    ]]];
      else
        candidateRight = sensitivitySolution->meshPoints[sensitivitySolution->nMeshPoints - 1];/*tEnd*/
      if((candidateLeft <= gammas[i]) && (gammas[i] < candidateRight)){
        dataPointsSegmentsIndices[i] = candidateIndex;
        prevFoundIndex = candidateIndex;
        found = DDEM_TRUE;
      }else{
        candidateIndex += 1;
				if (candidateIndex >= nImportantLambdas){
					if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED)
						cout << "(ddemParameterEstimator::setDataPointsSegmentsIndices)\n" <<
										" some data out of range i = " << i << "\n";
          error = DDEM_TRUE;
				}
      }
    }/* while */
		if(error)	break;
  }/* for */
	if(error == DDEM_TRUE)
		return DDEM_ERR_SENS_DATA_OUT_OF_RANGE;
	else
		return DDEM_SUCCESS;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/************************************************************/
double ddem_theta_0(double tRight, double tLeft, double t)
/* theta_0(tLeft) = 1.0 theta_0(tRight) = 0.0 */
/* theta_0'(tLeft) = 0.0 theta_0'(tRight) = 0.0 */
/* Ideally, we also need */
/* theta_0''(tLeft) = 0.0 theta_0''(tRight) = 0.0 */
{
	double Dt = tRight - tLeft;
	double a0, a1, a2, a3;
	a0 = 0.0;
	a1 = -1.0 / Dt;
	a2 = 1.0 / (Dt * Dt);
	a3 = 2.0 /(Dt * Dt * Dt);
	double theta_t;
	theta_t = ((a3 * (t - tRight) + a2) * (t - tLeft) + a1)
						* (t - tRight) + a0;
	return theta_t;
}
/************************************************************/
double ddem_theta_0_p(double tRight, double tLeft, double t)
{
	double Dt = tRight - tLeft;
	double  a1, a2, a3;
	a1 = -1.0 / Dt;
	a2 = 1.0 / (Dt * Dt);
	a3 = 2.0 /(Dt * Dt * Dt);
	double theta_t_p;
	double temp = (t - tLeft) + (t - tRight);
	theta_t_p = a3 * (t - tRight) * ((t - tLeft) + temp)
						+ a2 * temp + a1;
	return theta_t_p;
}

/************************************************************/
#define YSJAC(I,J) ysJac[(I)*l + (J)]
int ddemParameterEstimator::simpleEvaluate_(ddemSensitivityAnalyzer* sensitivityAnalyzer,
																				 int nGammas,
                                         double* gammas,
                                         int nYs,
                                         int* gammasIndices,
                                         int* YsIndices,
                                         /*int* dataPointsSegmentsIndices,*/
                                         int l,
                                         double* p,
                                         double* ys,
                                         double* ysJac,
                                         int flag,
																				 ddemSensitivitySolution* sensitivitySolution,
																				 int nImportantLambdas,
																				 int* importantLambdasIndices,
																				 ddemSimulator* simulator,
																				 int needSensitivity[],
																				 ddemSystem* baseSystem,
 																				 double tEnd,
																				 double tLeft)
{
int retCodeTemp;
if(1){/*GHGHGH*/
  int* dataPointsSegmentsIndices;
  dataPointsSegmentsIndices = new int[nGammas];

	retCodeTemp = setDataPointsSegmentsIndices_(sensitivitySolution,
																				 nImportantLambdas,
																				 importantLambdasIndices,
																				 gammas,
                                         nGammas,
                                         dataPointsSegmentsIndices);
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	retCodeTemp = 
	     configPreservEvaluate_(sensitivityAnalyzer,
												nGammas,
                        gammas,
                        nYs,
                        gammasIndices,
                        YsIndices,
                        dataPointsSegmentsIndices,
                        l,
                        p,
                        ys,
                        ysJac,
                        flag,
												sensitivitySolution,
												nImportantLambdas,
												importantLambdasIndices,
												simulator,
												needSensitivity,
												baseSystem,
												tEnd,
												tLeft);
	delete dataPointsSegmentsIndices;
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
}else{/*GHGHGH*/
	for (int i = 0; i < nYs; i++){
		int gammaIndex = gammasIndices[i];
		double t = gammas[gammaIndex];

//		#define YSJAC(I,J) ysJac[(I)*l + (J)]

		if ( t < p[0]){ // dataPointsSegmentsIndices[gammasIndices[i]] == 0){
      ys[i] = 9.0 + t;
      YSJAC(i,0) = 0.0;
    }else if ( t < 2 * p[0]){//(dataPointsSegmentsIndices[gammasIndices[i]] == 1){
      //ys[i] = .5*t*t+(9.0-p[0]+1.0)*t+.5*p[0]*p[0]-9.0*p[0]+9.0;
      //YSJAC(i,0) = -t+p[0]-9.0;
      ys[i] = 2*t-p[0]+9.0;
      YSJAC(i,0) = -1.0;
    }
    else if ( t < 3 * p[0]){// dataPointsSegmentsIndices[gammasIndices[i]] == 2 
      ys[i] = 3*t-3*p[0]+9.0;
      YSJAC(i,0) = -3;
    }
    else if ( t < 4 * p[0]){// dataPointsSegmentsIndices[gammasIndices[i]] == 2 
      ys[i] = 4*t-6*p[0]+9.0;
      YSJAC(i,0) = -6;
    }
    else if ( t < 5 * p[0]){// dataPointsSegmentsIndices[gammasIndices[i]] == 2 
      ys[i] = 5*t-10*p[0]+9.0;
      YSJAC(i,0) = -10;
    }
    else if ( t < 6 * p[0]){// dataPointsSegmentsIndices[gammasIndices[i]] == 2 
      ys[i] = 6*t-15*p[0]+9.0;
      YSJAC(i,0) = -15;
    }
    else if ( t < 7 * p[0]){// dataPointsSegmentsIndices[gammasIndices[i]] == 2 
      ys[i] = 7*t-21*p[0]+9.0;
      YSJAC(i,0) = -21;
    }
    else if ( t < 8 * p[0]){// dataPointsSegmentsIndices[gammasIndices[i]] == 2 
      ys[i] = 8*t-28*p[0]+9.0;
      YSJAC(i,0) = -28;
    }
		//printf("from hard t = %f      y = %.15f \n", t, ys[i]);
	}/*for*/
}/*GHGHGH*/
return DDEM_SUCCESS;
}
/************************************************************/
/************************************************************/
int ddemParameterEstimator::configPreservEvaluate_(ddemSensitivityAnalyzer* sensitivityAnalyzer,
																				 int nGammas,
                                         double* gammas,
                                         int nYs,
                                         int* gammasIndices,
                                         int* YsIndices,
                                         int* dataPointsSegmentsIndices,
                                         int l,
                                         double* p,
                                         double* ys,
                                         double* ysJac,
                                         int flag,
																				 ddemSensitivitySolution* sensitivitySolution,
																				 int nImportantLambdas,
																				 int* importantLambdasIndices,
																				 ddemSimulator* simulator,
																				 int needSensitivity[],
																				 ddemSystem* baseSystem,
																				 double tEnd,
																				 double tLeft)
{
//	#define YSJAC(I,J) ysJac[(I) * l + (J)]

	int retCodeTemp;
  for (int i = 0; i < nYs; i++){
		int gammaIndex = gammasIndices[i];
		//cout << "gammaIndex = " << gammaIndex << endl;
    double t = gammas[gammaIndex];
		//cout << " t = " << t << endl;
		//cout << "dataPointsSegmentsIndices[gammaIndex] = " << dataPointsSegmentsIndices[gammaIndex] << endl;
		if(dataPointsSegmentsIndices[gammaIndex] >= nImportantLambdas){/* tEnd <= lambda */
			//cout << "YES                      YES" << endl;
			ys[i] = 0.0;
			for(int k = 0; k < l; k++)
				YSJAC(i, k) = 0.0;
			if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED){
				cout << "<<<<<< SPECIAL SITUATION in ddemParameterEstimator::configPreservEvaluate >>>>>>> \n";
				cout << "tEnd <= lambda \n";
			}
		}else{/* lambda <= tEnd */
			int ImportantLeftMeshIndex = sensitivitySolution->lambdasIndices[
		                   importantLambdasIndices[
			                dataPointsSegmentsIndices[gammaIndex]
				              ]];
			//cout << "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX = " << << endl; 
			//cout << "ImportantLeftMeshIndex = " << ImportantLeftMeshIndex << endl; 
			double tImportantLeft = sensitivitySolution->meshPoints[ImportantLeftMeshIndex];
			double tImportantRight;
			int ImportantRightMeshIndex;
			if(dataPointsSegmentsIndices[gammaIndex] + 1 < nImportantLambdas)
				ImportantRightMeshIndex = sensitivitySolution->lambdasIndices[
					             importantLambdasIndices[
						          dataPointsSegmentsIndices[gammaIndex] + 1
							        ]]; 
			else
				ImportantRightMeshIndex = sensitivitySolution->nMeshPoints - 1;
			//cout << "ImportantRightMeshIndex = " << ImportantRightMeshIndex << endl; 
			tImportantRight = sensitivitySolution->meshPoints[ImportantRightMeshIndex];

			int betweenMeshPointsIndex;
			if(t < tImportantLeft)
				betweenMeshPointsIndex = ImportantLeftMeshIndex;
			else if( t >= tImportantRight)
				betweenMeshPointsIndex = ImportantRightMeshIndex - 1;
			else/* tImportantLeft <= t < tImportantRight*/
				for(int k = ImportantLeftMeshIndex; k < ImportantRightMeshIndex; k++)
					if(t < sensitivitySolution->meshPoints[k + 1]){
						betweenMeshPointsIndex = k;
						break;
					}
			//cout << "sensitivitySolution->nMeshPoints = " << sensitivitySolution->nMeshPoints << endl; 
			//cout << "ImportantLeftMeshIndex = " << ImportantLeftMeshIndex << endl; 
			//cout << "ImportantRightMeshIndex = " << ImportantRightMeshIndex << endl; 
			//cout << "betweenMeshPointsIndex = " << betweenMeshPointsIndex << endl;
			//cout << "t = " << t << endl;
			//cout << "sensitivitySolution->meshPoints[k] = " << sensitivitySolution->meshPoints[betweenMeshPointsIndex] << endl; 
			//cout << "sensitivitySolution->meshPoints[k + 1] = " << sensitivitySolution->meshPoints[betweenMeshPointsIndex + 1] << endl; 
			double* y;
			y = new double[sensitivitySolution->getNVariables()];
			//cout << "sensitivitySolution->getNVariables() = " << sensitivitySolution->getNVariables() << endl;
			retCodeTemp = 
			simulator->evaluateSolution2(t, y,
																sensitivitySolution,
																DDEM_INTERP,
																betweenMeshPointsIndex);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
			int yIndex = YsIndices[i];
			//cout << "yIndex = " << yIndex << endl;
			double lambda = tImportantLeft;
			double theta_0_lambda;
			if(lambda < tLeft){
				ys[i] = y[yIndex];
			}else{ /* tLeft < lambda <= tEnd */
				theta_0_lambda = ddem_theta_0(tEnd, tLeft, lambda); 
				ys[i] = y[yIndex] * theta_0_lambda;
				if(DDEM_TRACE_PARAMETER_ESTIMATOR_DETAILED){
					cout << "<<<<<< SPECIAL SITUATION in ddemParameterEstimator::configPreservEvaluate >>>>>>> \n";
					cout << "tLeft < lambda <= tEnd \n";
				}
			}
			int nSCounter = 0;
			int m = baseSystem->getNVariables();
			for(int k = 0; k < l; k++){
				if(needSensitivity[k] == DDEM_TRUE){
					YSJAC(i, k) = y[m * (nSCounter + 1) + yIndex]; 
					//cout << "XXXXXXXXXXXXXXXXXX = " << m * (nSCounter + 1) + yIndex << endl;
					nSCounter+=1;
				}else
					YSJAC(i, k) = 0.0;
				if(lambda > tLeft){ /* tLeft < lambda <= tEnd */
					//cout << "XXXXXYESXXXXXXXXXXXXX = " << endl;
					double par_lambda_par_p_k = sensitivitySolution->
					                             lambdasJac[
						                            importantLambdasIndices[
																					dataPointsSegmentsIndices[gammaIndex]]][k];
					YSJAC(i, k) = ddem_theta_0_p(tEnd, tLeft, lambda) *
												par_lambda_par_p_k *
												y[yIndex] + 
												theta_0_lambda * YSJAC(i, k);
				}/*if*/
			}/* for k */
			delete y;
		}/*if*/
	}/* for i */
	return DDEM_SUCCESS;
}


