#ifndef __DDEM_PARAMETER_ESTIMATOR
#define __DDEM_PARAMETER_ESTIMATOR
/************************************************************/
#include <nag.h>

#include "ddem_system.h"
#include "ddem_simulator.h"
#include "ddem_sensitivity_analyzer.h"
#include "ddem_data.h"
#include "ddem_constraints.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
struct ddemParameterEstimationStatistics: public ddemSimulationStatistics{

	int nReferencedParameterValues;
	double finalObjectiveValue;

	ddemParameterEstimationStatistics();
};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
class ddemParameterEstimator;
class ddemData;
class ddemConstraints;
struct ddemHelperComm{
		ddemParameterEstimationStatistics* statistics;

		double relTolerance;
		double* absTolerance;
		double vanishingConst;

		ddemParameterEstimator* parameterEstimator;
		int first;
		ddemSensitivitySolution* lastSensitivitySolution;
		double* lastParameters;
		double tLeft;
		int peSelect;
		double tEnd;
		void* commWithUFuncs;
		ddemSimulator* simulator;
    ddemSystem* system;
    ddemData* data;
		ddemConstraints* constraints;
    ddemSensitivityAnalyzer* sensitivityAnalyzer;
    int* C012;
    int nC012;
    int* dataPointsSegmentsIndices;

		int nImportantLambdas;
	  int* importantLambdasIndices;;

};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
#define DDEM_PE_VERY_SIMPLE  1 /* No constraints are added for important lambdas. */
														/* Jacobians of the objective function are computed by */
														/* divided differences. */
#define DDEM_PE_SIMPLE  2 /* No constraints are added for important lambdas. */
#define DDEM_PE_NEW     3	/* Constraints are added, but function evaluation is simple */
#define DDEM_PE_NEW_CE  4 /* Constarints are added, and function evaluation */
												  /* is done using continuouse extensions when a constraint is violated.*/	
class ddemParameterEstimator{
public:
    ddemParameterEstimator(){};
   ~ddemParameterEstimator(){};
    int EstimateParameters(ddemSimulator* simulator,
														ddemSystem* system,
                            ddemData* data,
                            ddemConstraints* constraints,
                            double parameters[],
														double relTolerance,
														double* absTolerance,
														double vanishingConst,
														ddemParameterEstimationStatistics* statistics,
														int peSelect,/* DDEM_PE_SIMPLE, DDEM_PE_NEW*/
														void* commWithUFuncs = NULL);/**/
    int provideSensitivities_(ddemSimulator* simulator,
															ddemSystem* system, double tEnd,
                              double parameters[],
															double relTolerance,
															double* absTolerance,
															double vanishingConst,
															int needSensitivity[],
                              //ddemSensitivitySolution*& sensitivitySolution,
															//int& nImportantLambdas,
															//int*& importantLambdasIndices,
															void* commWithUFuncs,
															ddemHelperComm* helperComm);
  friend void subfunctionsCalculator(Integer m, Integer n, double p[],
    				                     double f[], double fjac[], Integer tdfjac,
		                             Nag_Comm *comm);

private:
    int setC012Constraints_(int& nC012, int* C012,
                                        int* blC012GammaIndices,
                                        int* buC012GammaIndices,
                                        double* bl, double * bu,
                                        ddemData* data,
                                        ddemConstraints* constraints,
                                        int* dataPointsSegmentsIndices,
                                        int nImportantLambdas);

    int setDataPointsSegmentsIndices_(ddemSensitivitySolution* sensitivitySolution,
																			int nImportantLambdas,
																			int* importantLambdasIndices,
																		  double* gammas, int nGammas,
																			int* dataPointsSegmentsIndices);

		int simpleEvaluate_(ddemSensitivityAnalyzer* sensitivityAnalyzer,
																				 int nGammas,
                                         double* gammas,
                                         int nYs,
                                         int* gammasIndices,
                                         int* YsIndices,
                                         /*int* dataPointsSegmentsIndices,*/
                                         int l,
                                         double* p,
                                         double* ys,
                                         double* ysJac,
                                         int flag,
																				 ddemSensitivitySolution* sensitivitySolution,
																				 int nImportantLambdas,
																				 int* importantLambdasIndices,
																				 ddemSimulator* simulator,
																				 int needSensitivity[],
																				 ddemSystem* baseSystem,
																				 double tEnd,
																				 double tLeft);
	  int configPreservEvaluate_(ddemSensitivityAnalyzer* sensitivityAnalyzer,
																						 int nGammas,
                                             double* gammas,
                                             int nYs,
                                             int* gammasIndices,
                                             int* YsIndices,
                                             int* dataPointsSegmentsIndices,
                                             int l,
                                             double* p,
                                             double* ys,
                                             double* ysJac,
                                             int flag,
																						 ddemSensitivitySolution* sensitivitySolution,
																						 int nImportantLambdas,
																						 int* importantLambdasIndices,
																						 ddemSimulator* simulator,
																						 int needSensitivity[],
																						 ddemSystem* baseSystem,
																						 double tEnd,
																						 double tLeft);

		void setImportantLambdasIndices_(ddemSensitivitySolution* sensitivitySolution,
																		 int& nImportantLambdas,
																		 int*& importantLambdasIndices,
																						 int importanceLevel);

};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void subfunctionsCalculator(Integer m, Integer n, double p[],
    				                     double f[], double fjac[], Integer tdfjac,
		                             Nag_Comm *comm);
void constraintsCalculator(Integer n, Integer ncnlin, Integer needc[],
		                            double p[], double conf[], double cjac[],
		                            Nag_Comm *comm);
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
double ddem_theta_0(double tRight, double tLeft, double t);
double ddem_theta_0_p(double tRight, double tLeft, double t);
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
#endif /* __DDEM_PARAMETER_ESTIMATOR*/
