/* DOUBLE CHECKED = YES*/
#include <iostream>
using namespace std;

#include "ddem_sensitivity_analyzer.h"
#include "ddem_fsens_system.h"

#include "ddem_trace.h"
#include "ddem_error_codes.h"
#include "ddem_sol_out.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
int ddemSensParameterSelector::setNVariablesBase(int nVariablesBase)
{
	if(nVariablesBase_ == DDEM_UNDEFINED){
		nVariablesBase_ = nVariablesBase;
		return DDEM_SUCCESS;
	}else
		return DDEM_FAILURE;

}
/************************************************************/
int ddemSensParameterSelector::getNVariablesBase()
{
	return nVariablesBase_;
}
/************************************************************/
int ddemSensParameterSelector::getParameterPlaceInSelection(int parameterIndex)
{
	if((parameterIndex >=0) &&(parameterIndex < nParameters_))
		return selParametersRevIndices_[parameterIndex];
	else
		return DDEM_UNDEFINED;
}
/************************************************************/
int ddemSensParameterSelector::setParametersPlacesInSelection(int nParameters, int nS, int* sIndices)
{
	nParameters_ = nParameters;
	if(selParametersRevIndices_ == NULL){
		selParametersRevIndices_ = new int[nParameters];
		for(int j=0; j < nParameters; j++)
			selParametersRevIndices_[j] = DDEM_UNDEFINED;
		for(int i=0; i < nS; i++)
			selParametersRevIndices_[sIndices[i]] = i;
		return DDEM_SUCCESS;
	}else
		return DDEM_FAILURE; 
}
/************************************************************/
ddemSensParameterSelector::ddemSensParameterSelector()
{
	selParametersRevIndices_ = NULL;

	nVariablesBase_ = DDEM_UNDEFINED;
	nParameters_ = DDEM_UNDEFINED;
}
/************************************************************/
ddemSensParameterSelector::~ddemSensParameterSelector()
{
	if(selParametersRevIndices_ != NULL)
		delete selParametersRevIndices_;

}
/************************************************************/
ddemSensitivitySolution::ddemSensitivitySolution()
{
	lambdasJac = NULL;

	historyStartsLambdasJac = NULL;

	isAllocatedExtra_ = DDEM_FALSE;

}
/************************************************************/
ddemSensitivitySolution::~ddemSensitivitySolution()
{
	releaseExtraMemories();
}
/************************************************************/
int ddemSensitivitySolution::isAllocatedExtra()
{
	return isAllocatedExtra_;
}
/************************************************************/
void ddemSensitivitySolution::allocateExtraMemories(int nSelectedParameters,
																										int nHistorySegments)
{
	if(isAllocatedExtra_ == DDEM_TRUE){
		if(DDEM_TRACE_SENSITIVITY_ANALYZER)
			cout << "ddemSensitivitySolution::allocateExtraMemories can't be called before more than once"
				<<"\n";
		return ;
	}
	lambdasJac = new double*[DDEM_MAX_SPACE_LAMBDAS];
	for( int i=0; i < DDEM_MAX_SPACE_LAMBDAS; i++)
		lambdasJac[i] = new double[nSelectedParameters];

	if(nHistorySegments > 1){
		historyStartsLambdasJac = new double*[nHistorySegments - 1];
		for( int j=0; j < nHistorySegments - 1; j++)
			historyStartsLambdasJac[j] = new double[nSelectedParameters];
	}
	
	isAllocatedExtra_ = DDEM_TRUE;

	/*nSelectedParameters_ = nSelectedParameters;*/
}
/************************************************************/
void ddemSensitivitySolution::releaseExtraMemories()
{
	if(isAllocatedExtra_ == DDEM_FALSE){
		if(DDEM_TRACE_SENSITIVITY_ANALYZER)
			cout << "ddemSensitivitySolution::releaseExtraMemories"
				<< "can't be called before allocateExtraMemories()"
				<<"\n";
		return;
	}
	for( int i=0; i < DDEM_MAX_SPACE_LAMBDAS; i++)
		delete lambdasJac[i];
	delete lambdasJac; lambdasJac = NULL;

	if(historyStartsLambdasJac != NULL){
		for( int j=0; j < getNHistorySegments() - 1; j++)
			delete historyStartsLambdasJac[j];
		delete historyStartsLambdasJac; historyStartsLambdasJac = NULL;
	}

	isAllocatedExtra_ = DDEM_FALSE;
}
/************************************************************/
void needSensitivity2Indices(int nParameters, int* needSensitivity,
														 int* sIndices, int& nS)
{
	nS = 0;
	for(int i=0; i < nParameters; i++)
		if(needSensitivity[i] == DDEM_TRUE){
			sIndices[nS] = i;				
			nS += 1;
		}
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
ddemSensitivityAnalyzer::ddemSensitivityAnalyzer()
{
}
/************************************************************/
ddemSensitivityAnalyzer::~ddemSensitivityAnalyzer()
{
}
/************************************************************/
int ddemSensitivityAnalyzer::initializeIndices_(
													ddemSystem* baseSystem,
													int needSensitivity[],
													int& nS,
													int*& sIndices,
													ddemSensParameterSelector* sensParameterSelector)
{
	sIndices = new int[baseSystem->getNParameters()];
	needSensitivity2Indices(baseSystem->getNParameters(), needSensitivity,
														 sIndices, nS);


	sensParameterSelector->setNVariablesBase(baseSystem->getNVariables());
	sensParameterSelector->setParametersPlacesInSelection(baseSystem->getNParameters(), nS, sIndices);

	return DDEM_SUCCESS;
}
/************************************************************/
int ddemSensitivityAnalyzer::computeSensitivities(ddemSimulator* simulator,
													ddemSystem* baseSystem, double tEnd,
                          double* parameters,
													double relToleranceBase,
													double* absToleranceBase,
													double vanishingConst,
													int* needSensitivity,
													void* commWithUFuncs,
                          ddemSensitivitySolution* sensitivitySolution)
{
	int retCodeTemp;
	if(DDEM_TRACE_SENSITIVITY_ANALYZER){
		cout << "parameter = " << parameters[0] << "\n";
	}

	int* sIndices;
	int nS;
	initializeIndices_(baseSystem, needSensitivity, nS, sIndices, sensitivitySolution);
	sensitivitySolution->allocateExtraMemories(nS, baseSystem->getNHistorySegments());



	ddemFSensSystem* fSensSystem;
	fSensSystem = new ddemFSensSystem(baseSystem,
																	nS, sIndices);
	retCodeTemp =
	fSensSystem->compPartialsT0(parameters,
												sensitivitySolution->lambdasJac[0],
												commWithUFuncs);
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	for(int i=0; i < fSensSystem->getNHistorySegments() - 1; i++){
			double* par_lambda_i_par_p = sensitivitySolution->historyStartsLambdasJac[i];
			retCodeTemp =
			fSensSystem->evalPartialsHistoryStart(i + 1,
															parameters,
															par_lambda_i_par_p,
															commWithUFuncs);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
	}
	ddemViolation* violated;
	violated = new ddemViolation(fSensSystem->getNu() + fSensSystem->getOmega());
	int atEnd = DDEM_FALSE;
	int iterCounter = 0;
	/* setting error tolerances */
	double relTolerance = relToleranceBase;
	double* absTolerance;
	absTolerance = new double[fSensSystem->getNVariables()];
	int nBaseSystem = baseSystem->getNVariables();
	for(int i = 0; i < nBaseSystem; i++)
		absTolerance[i] = absToleranceBase[i];
	for(int k=0; k < nS; k++){
		int parameterIndex = sIndices[k];
		double parameterSelected = parameters[parameterIndex];
		double pScale = ddem_max(ddem_fabs(parameterSelected), 1e-40);
		for(int i =0; i < nBaseSystem; i++)
			absTolerance[(1 + k) * nBaseSystem + i] = absToleranceBase[i] / pScale;
	}
	if(baseSystem->getOmega() > 0)
		for(int i =0; i < nBaseSystem; i++)
			absTolerance[(1 + nS) * nBaseSystem + i] = absToleranceBase[i];
	/*-------------------------*/
	do{
		int retCodeSim =
    simulator ->simulateWhileContinuous(fSensSystem, tEnd,
                                parameters,
																relTolerance,
																absTolerance,
																vanishingConst,
	                              commWithUFuncs,
  	                            1,/*discontinuityLevel*/
  	                            /* 2 if second order sensitivities */
																violated,
																atEnd,
																sensitivitySolution);
		if(retCodeSim != DDEM_SUCCESS)
			return retCodeSim;
		if(atEnd == DDEM_TRUE)
			break;
		/* compute the value of jump and set new staring values */
		double t;
		double* y;
		double* ypL;
		double* ypR;

		t = sensitivitySolution->meshPoints[sensitivitySolution->nMeshPoints - 1];
		y = sensitivitySolution->ys[sensitivitySolution->nMeshPoints - 1];
		
		ypL = new double[fSensSystem->getNVariables()];
		retCodeTemp =
		simulator->evaluateSolution2(t, ypL, sensitivitySolution, DDEM_DERIV_INTERP,
																sensitivitySolution->nMeshPoints - 2);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;

		double par_alpha_disc_par_t;
		double* par_alpha_disc_par_y;
		double* par_alpha_disc_par_p;


		int m = baseSystem->getNVariables();
		par_alpha_disc_par_y = new double[m];
		par_alpha_disc_par_p = new double[nS];
		retCodeTemp =
		fSensSystem->evalPartialsDelayArgumentsBaseSystem(t, y,
																parameters,
																violated->violAlphaIndex,
																par_alpha_disc_par_t,
																par_alpha_disc_par_y,
																par_alpha_disc_par_p,
																commWithUFuncs);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;

		double* par_lambda_rPlus1_par_p = sensitivitySolution->lambdasJac[
														sensitivitySolution->nLambdas - 1];

		double* par_lambda_r_par_p;

		int prevLambdaIndex = violated->baseIndices[violated->violAlphaIndex];
		if(prevLambdaIndex >= 0)
			par_lambda_r_par_p = sensitivitySolution->lambdasJac[prevLambdaIndex];
		else
			par_lambda_r_par_p = sensitivitySolution->historyStartsLambdasJac[-prevLambdaIndex - 1];
		/*compute ypR*/
		ypR = new double[m];
		int omega = baseSystem->getOmega();
		int nu = baseSystem->getNu();
		double* zBaseSystem;
		zBaseSystem = new double[(nu + omega) * m];
		/*compute zBaseSystem*/
		retCodeTemp =
		compZBaseSystemAtDisc_(parameters,
														simulator,
														baseSystem,
														fSensSystem->getNVariables(),
														violated,
														sensitivitySolution,
														zBaseSystem,
														commWithUFuncs);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;
		retCodeTemp =
		baseSystem->evalF(t,
											y,
											zBaseSystem,
											parameters,
											ypR,
											commWithUFuncs);
		if(retCodeTemp != DDEM_SUCCESS)
			return retCodeTemp;
		/*compute par_lambda_rPlus1_par_p*/
		for(int i=0; i < nS; i++){
			double* par_y_par_p_i;
			par_y_par_p_i = &y[(1 + i) * m];
			double accN;
			accN = 0.0;
			for(int j=0; j < m; j++)
				accN+= par_alpha_disc_par_y[j] * par_y_par_p_i[j];
			accN+= par_alpha_disc_par_p[i];
			accN-= par_lambda_r_par_p[i];
			double accD =0;
			accD = 0.0;
			for(int j=0; j < m; j++)
				accD+= par_alpha_disc_par_y[j] * ypL[j];
			accD+= par_alpha_disc_par_t;
			if(accD == 0.0){
				if(DDEM_TRACE_SENSITIVITY_ANALYZER)
					cout << "Error in ddemSensitivityAnalyzer::computeSensitivities\n" <<
									"-----vanishing denominator------";
				return DDEM_ERR_SENS_VANISHING_DENOMINATOR;
			}else
				par_lambda_rPlus1_par_p[i] = - accN / accD;
			for(int j=0; j < m; j++)
				par_y_par_p_i[j] = par_y_par_p_i[j] + (ypL[j] - ypR[j])
												* par_lambda_rPlus1_par_p[i]; 
		}
		if(omega > 0){
			double* x = &y[(1 + nS) * m];
			for(int i=0; i < m; i++)
				x[i] = ypR[i];
		}
		/* set the integ*/
		iterCounter+=1;
		if(t  >= tEnd)
			atEnd = DDEM_TRUE;
		delete ypR;
		delete zBaseSystem;
		delete ypL;
		delete par_alpha_disc_par_y;
		delete par_alpha_disc_par_p;
	} while (!atEnd);
	if(DDEM_TRACE_SENSITIVITY_ANALYZER){
		cout << "nMeshPoints = " << sensitivitySolution->nMeshPoints << "\n";
		for(int jj=0; jj < sensitivitySolution->nLambdas; jj++){
			int ii = sensitivitySolution->lambdasIndices[jj];
			cout <<"DiscPoint["<<jj<<"] = " << sensitivitySolution->meshPoints[ii] <<
			"  y["<<ii<<"][0] = " << sensitivitySolution->ys[ii][0] <<"  "
			/*"  y["<<ii<<"][1] = " << sensitivitySolution->ys[ii][1] <<"  "*/
			"  y["<<ii<<"][1] = " << sensitivitySolution->ys[ii][1] <<"\n";
		}
	}
	delete absTolerance;
	delete violated;
	delete fSensSystem;
	delete sIndices;
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemSensitivityAnalyzer::compZBaseSystemAtDisc_(double* parameters,
														ddemSimulator* simulator,
														ddemSystem* baseSystem,
														int sensNVariables,
														ddemViolation* violated,
														ddemSimulationSolution* simulationSolution,
														double* zBaseSystem,
														void* commWithUFuncs)
{
	int retCodeTemp;

	double* alpha;
	int m = baseSystem->getNVariables();
	int nu = baseSystem->getNu();
	int omega = baseSystem->getOmega();


	alpha = violated->alphasAtTn;
	double* ySens;
	ySens = new double[sensNVariables];
	for(int k=0; k < nu ; k++){
		if(violated->baseIndices[k] >= 0){
			retCodeTemp =
			simulator->evaluateSolution(alpha[k], ySens,
																	simulationSolution, DDEM_INTERP);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
		}else{
			retCodeTemp =
			baseSystem->evalHistory(-violated->baseIndices[k], alpha[k], parameters,
																			ySens, commWithUFuncs);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
		}
		for(int i=0; i < m; i++)
			zBaseSystem[k*m + i] = ySens[i];
	}

	for(int k=nu; k < nu + omega; k++){
		if(violated->baseIndices[k] >= 0){
			retCodeTemp =
			simulator->evaluateSolution(alpha[k], ySens,
																	simulationSolution, DDEM_DERIV_INTERP);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
		}else{
			retCodeTemp =
			baseSystem->evalDerivHistory(-violated->baseIndices[k], alpha[k], parameters,
																			ySens, commWithUFuncs);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
		}
		for(int i=0; i < m; i++)
			zBaseSystem[k*m + i] = ySens[i];
	}
	delete ySens;
	return DDEM_SUCCESS;
}
/************************************************************/
ddemSensitivitySolutionFD::ddemSensitivitySolutionFD()
{
	isAllocated_ = DDEM_FALSE;
	ys = NULL;
	ts = NULL;
}
ddemSensitivitySolutionFD::~ddemSensitivitySolutionFD()
{
	if(isAllocated_ == DDEM_TRUE)
		releaseMemories_();
}
int ddemSensitivitySolutionFD::allocateMemories(int nVariablesBase, int nS, int nPoints)
{
	if(isAllocated_ == DDEM_TRUE){
		cout << "ddemSensitivitySolutionFD::allocateMemories was already called"
				<<"\n";
		return DDEM_FAILURE;
	}else{
		ts = new double[nPoints];
		ys = new double*[nPoints];
		for(int i=0; i < nPoints; i++)
			ys[i] = new double[(1 + nS) * nVariablesBase];
		nPoints_ = nPoints;
		nS_ = nS;

		isAllocated_ = DDEM_TRUE;
		return DDEM_SUCCESS;
	}
}
int ddemSensitivitySolutionFD::releaseMemories_()
{
	if(isAllocated_ == DDEM_TRUE){
		for(int i=0; i < nPoints_; i++)
			delete ys[i];
		delete ys; ys = NULL;
		delete ts; ts = NULL;
	}
	return DDEM_SUCCESS;
}

/************************************************************/
int ddemSensitivityAnalyzer::computeSensitivitiesFD(ddemSimulator* simulator,
															ddemSystem* baseSystem, double tEnd,
                              double* parameters,
															double deltaP,
															double relToleranceBase,
															double* absToleranceBase,
															double vanishingConst,
															int needSensitivity[],
															int nOutPoints,
															void* commWithUFuncs,
                              ddemSensitivitySolutionFD* sensitivitySolutionFD)
{
	int* sIndices;
	int nS;
	initializeIndices_(baseSystem, needSensitivity, nS, sIndices, sensitivitySolutionFD);
	
	int m = baseSystem->getNVariables();

	sensitivitySolutionFD->allocateMemories(m, nS, nOutPoints);

	ddemSimulationSolution* simulationSolution;
	simulationSolution = new ddemSimulationSolution();
	/*--------------*/
	ddemUniformStepper* uniformStepper; 
	uniformStepper = new ddemUniformStepper();
	
	simulator->simulate(baseSystem,
											tEnd,
											parameters,
											relToleranceBase,
											absToleranceBase,
											vanishingConst,
											commWithUFuncs,
											simulationSolution);


	uniformStepper->initialize(simulationSolution->meshPoints[0], tEnd, nOutPoints);

	double t;
	for(int k =0; k < nOutPoints; k++)
		if(uniformStepper->get(k, t) == DDEM_SUCCESS)
			sensitivitySolutionFD->ts[k] = t;

	double* y;
	y = new double[m];
	

	for(int k =0; k < nOutPoints; k++){
		simulator->evaluateSolution(sensitivitySolutionFD->ts[k], y, simulationSolution, DDEM_INTERP, DDEM_UNKNOWN);
		for(int a = 0; a < m; a++)
			sensitivitySolutionFD->ys[k][a] = y[a];
	}

	double* parametersDeltaP;
	int nParameters;
	nParameters = baseSystem->getNParameters();
	parametersDeltaP = new double[nParameters];
	for(int j=0; j < nParameters; j++)
		parametersDeltaP[j] = parameters[j];

	double* yDeltaP;
	yDeltaP = new double[m];


	for(int i=0; i < nS; i++){
		if( i > 0)/* restore from the last change*/ 
			parametersDeltaP[sIndices[i - 1]] = parameters[sIndices[i - 1]];
		parametersDeltaP[sIndices[i]] = parametersDeltaP[sIndices[i]] + deltaP;
		ddemSimulationSolution* simulationSolutionDeltaP;
		simulationSolutionDeltaP = new ddemSimulationSolution();
		simulator->simulate(baseSystem,
												tEnd,
												parametersDeltaP,
												relToleranceBase,
												absToleranceBase,
												vanishingConst,
												commWithUFuncs,
												simulationSolutionDeltaP);
		/*&&&&&&&&&&&&&&&&*/
		for(int k =0; k < nOutPoints; k++){
			int retCodeEval =
			simulator->evaluateSolution(sensitivitySolutionFD->ts[k], yDeltaP, simulationSolutionDeltaP, DDEM_INTERP, DDEM_UNKNOWN);
			if(retCodeEval != DDEM_SUCCESS){
				if(retCodeEval == DDEM_ERR_SIM_T_LESS_THAN_T0){
					int betweenMeshPointsIndex = 0;
					int betweenLambdasIndex;
					baseSystem->locateBetweenHistoryStarts(sensitivitySolutionFD->ts[k], simulationSolutionDeltaP->historyStarts,
																					betweenLambdasIndex);
					baseSystem->evalHistory(betweenLambdasIndex, sensitivitySolutionFD->ts[k],
																								parametersDeltaP, yDeltaP, commWithUFuncs);
				}else
					return retCodeEval;
			}

			for(int a = 0; a < m; a++)
				sensitivitySolutionFD->ys[k][(1 + i) * m + a] = (yDeltaP[a] 
																													- sensitivitySolutionFD->ys[k][a]) / deltaP;
		}
		delete simulationSolutionDeltaP;
	}
	/*--------------*/
	delete yDeltaP;
	delete y;
	delete uniformStepper;
	delete simulationSolution;
	delete parametersDeltaP;
	delete sIndices;
	return DDEM_SUCCESS;

}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
