/* DOUBLE CHECKED = YES*/
#include <iostream>
using namespace std;

#include <math.h>

#include "ddem_simulator.h"
#include "ddem_min_root.h"

#include "ddem_trace.h"
#include "ddem_error_codes.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
ddemSimulationStatistics::ddemSimulationStatistics()
{
	nSteps = 0;
	nFails = 0;
	nFEvals = 0;

	nHistoryEvals = 0;
	nDelayEvals = 0;
	nRootFindings = 0;
	nInterpEvals = 0;

	nVanishingSteps = 0;
	nIteraions4VanishingSteps = 0;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
ddemSimulationSolution::ddemSimulationSolution()
{
	statistics = new ddemSimulationStatistics();
	
	nLambdas = 0;
	lambdasDiscDegrees = NULL;
	lambdasIndices = NULL;
	nMeshPoints = 0;
	meshPoints = NULL;
	ys = NULL;
	interpolants = NULL;
	interpStepSizes = NULL;

	historyStarts = NULL;
	historyStartsDiscDegrees = NULL;

	isAllocated_ = DDEM_FALSE;
}
/************************************************************/
ddemSimulationSolution::~ddemSimulationSolution()
{	if(isAllocated_ == DDEM_TRUE)
		releaseMemories();
	delete statistics; statistics = NULL;
}
/************************************************************/
int ddemSimulationSolution::getNVariables()
{
	if(isAllocated_  == DDEM_TRUE)
		return nVariables_;
	else{
		cout << "ddemSimulationSolution::getNVariables can't be called before "
			<< "ddemSimulationSolution::allocateMemories" << endl;
		return DDEM_UNDEFINED;
	}
}
/************************************************************/
int ddemSimulationSolution::getNHistorySegments()
{
	if(isAllocated_  == DDEM_TRUE)
		return nHistorySegments_;
	else{
		cout << "ddemSimulationSolution::getNHistorySegments can't be called before "
			<< "ddemSimulationSolution::allocateMemories" << endl;
		return DDEM_UNDEFINED;
	}
}
/************************************************************/
int ddemSimulationSolution::isAllocated()
{
	return isAllocated_;
}
/************************************************************/
int ddemSimulationSolution::allocateMemories(int nVariables, int nInterploatorCoefficens,
																						 int nHistorySegments)
{
	if(isAllocated_  == DDEM_TRUE){
		cout << "ddemSimulationSolution::allocateMemories can't be called more than once"
			<< endl;
		return DDEM_ERR_UNEXPECTED;
	}

	meshPoints = new double[DDEM_MAX_SPACE_MESH_POINTS];
	ys = new double*[DDEM_MAX_SPACE_MESH_POINTS];
	interpolants = new double**[DDEM_MAX_SPACE_MESH_POINTS];
	interpStepSizes = new double[DDEM_MAX_SPACE_MESH_POINTS];
	for(int i=0; i < DDEM_MAX_SPACE_MESH_POINTS; i++){
		ys[i] = new double[nVariables];
		interpolants[i] = new double*[nVariables];
		for(int j=0; j < nVariables; j++)
			interpolants[i][j] = new double[nInterploatorCoefficens];
	}
	
	lambdasIndices = new int[DDEM_MAX_SPACE_LAMBDAS];
	lambdasDiscDegrees = new int[DDEM_MAX_SPACE_LAMBDAS];
		
	if(nHistorySegments > 1){
		historyStarts = new double[nHistorySegments - 1];
		historyStartsDiscDegrees = new int[nHistorySegments];
	}

	isAllocated_ = DDEM_TRUE;

	nVariables_ = nVariables;
	nHistorySegments_ = nHistorySegments;

	return DDEM_SUCCESS;
}
/************************************************************/
int ddemSimulationSolution::releaseMemories()
{
	if(isAllocated_ == DDEM_FALSE){
		cout << "ddemSimulationSolution::releaseMemories can't be called before allocateMemories()"
			<< endl;
		return DDEM_ERR_UNEXPECTED;
	}
	delete meshPoints;  meshPoints = NULL;
	delete interpStepSizes;  interpStepSizes = NULL;
	for(int i=0; i < DDEM_MAX_SPACE_MESH_POINTS; i++){
		delete ys[i];
		for(int j=0; j < nVariables_; j++)
			delete interpolants[i][j];
		delete interpolants[i];
	}
	delete 	ys; ys = NULL;
	delete interpolants; interpolants = NULL;

	delete lambdasIndices;  lambdasIndices = NULL;
	delete lambdasDiscDegrees;  lambdasDiscDegrees = NULL;

	if(historyStarts != NULL) {delete historyStarts; historyStarts = NULL;}
	if(historyStartsDiscDegrees != NULL) 
	{	delete historyStartsDiscDegrees;
		historyStartsDiscDegrees = NULL;
	}

	isAllocated_ = DDEM_FALSE;

	return DDEM_SUCCESS;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
ddemViolation::ddemViolation(int nDelays)
{
	before = new int[nDelays];
	tSource = new double[nDelays];
	t = new double[nDelays];
	tZero = new double[nDelays];
	violType = new int[nDelays];
	direction = new int[nDelays];
	baseIndices = new int[nDelays];
	alphasAtTn = new double[nDelays];
	alphasAtTRight = new double[nDelays];

	isVanishing = new int[nDelays];

	nDelays_ = nDelays;
}
/************************************************************/
void ddemViolation::clear()
{
	any = DDEM_FALSE;
	for(int i=0; i < nDelays_; i++)
		before[i] = DDEM_FALSE;
}
/************************************************************/
ddemViolation::~ddemViolation()
{
	delete before;
	delete tSource;
	delete t;
	delete tZero;
	delete violType;
	delete direction;
	delete baseIndices;
	delete alphasAtTn;
	delete alphasAtTRight;

	delete isVanishing;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
int simIVPfcn(int*n, double* x, double* y, double* yprime,
													 void* ivp2ddeCommVoid)
{
  ddemivp2ddeComm* ivp2ddemComm;
	ivp2ddemComm = (ddemivp2ddeComm*)ivp2ddeCommVoid;
	
	/*compute delay arguments*/
	int retCodeTemp;
	retCodeTemp =
		ivp2ddemComm->system->evalDelayArguments(*x/*double t*/,
																		y/*double* y*/,
																		ivp2ddemComm->parameters/*double* p*/,
																		ivp2ddemComm->alpha/*double* alpha*/,
																		ivp2ddemComm->commWithUFuncs/*void* comm*/);
	ivp2ddemComm->statistics->nDelayEvals += 1;
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	/*computation of z*/
	int nu = ivp2ddemComm->system->getNu();
	int m = *n;

	double zeta;
	if(nu > 0)
		for(int i=0; i < nu; i++){
			retCodeTemp =
					ivp2ddemComm->simulator->evaluateCExtension(*x, ivp2ddemComm->alpha[i],
																										&(ivp2ddemComm->z[i * m]),
																										i,
																										DDEM_INTERP);
			if(ivp2ddemComm->violated->isVanishing[i] == DDEM_TRUE){
				if(retCodeTemp == DDEM_ERR_SIM_REQUIRED_EXTENSION_IS_IMPOSSIBLE){
					if(fabs(*x - *(ivp2ddemComm->addTn)) > 1e-15){ //!= 0.0
						zeta = (ivp2ddemComm->alpha[i] - *(ivp2ddemComm->addTn))/
									(*x - *(ivp2ddemComm->addTn));
					}else zeta = 0.0;
					double* yAlpha = &(ivp2ddemComm->z[i * m]);
					for(int k = 0; k < m; k++)
						yAlpha[k] = (1.0 - zeta) * ivp2ddemComm->yTn[k] + zeta * y[k];
				}else{
					if(retCodeTemp != DDEM_SUCCESS)
						return retCodeTemp;
				}
			}
		}
	int omega = ivp2ddemComm->system->getOmega();
	if(omega > 0)
		for(int j=0; j < omega; j++){
			ivp2ddemComm->simulator->evaluateCExtension(*x, ivp2ddemComm->alpha[nu + j],
																			&(ivp2ddemComm->z[(nu + j) * m]),
																			nu + j,
																			DDEM_DERIV_INTERP);
		}
	/*call user defined F*/
	ivp2ddemComm->system->evalF
										((*x)/*double t*/, y/*double* y*/,
										 ivp2ddemComm->z/*double* z*/,
										 ivp2ddemComm->parameters/*double* p*/,
										 yprime/*double* f*/,
										 ivp2ddemComm->commWithUFuncs/*void* comm*/);
	ivp2ddemComm->statistics->nFEvals += 1;


  return DDEM_SUCCESS;
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void ddemDelay2g(double t, double* g, void* voidComm)
{
	ddemDelays2MinRootComm* comm;
	comm = (ddemDelays2MinRootComm*) voidComm;
	comm->IVP2DDE->interpolate(comm->nVariables,
														comm->tLeftMeshPoint,
														comm->yLeftMeshPoint,
														t, comm->y,
														comm->stepSize,
														comm->interpolant,
														DDEM_INTERP);
	comm->system->evalDelayArguments(t/*double t*/,
																		comm->y/*double* y*/,
																		comm->parameters/*double* p*/,
																		comm->alpha/*double* alpha*/,
																		comm->commWithUFuncs/*void* comm*/);
	comm->statistics->nDelayEvals += 1;

	for(int i=0; i < comm->ng; i++)
		g[i] = comm->alpha[comm->candidateActiveIndices[i]] - 
						comm->tZero[comm->candidateActiveIndices[i]];
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
int ddemSimulator::simulateWhileContinuous(ddemSystemInterface* system,
																						/*double* restartY,*/
                                  double tEnd,
                                  double* parameters,
																	double relTolerance,
																	double* absTolerance,
																	double vanishingConst,
																	void* commWithUFuncs,
                                  int discontinuityLevel,
																	ddemViolation* violated,
																	int& atEnd,
                                  ddemSimulationSolution* simulationSolution)
{
	simulating_ = DDEM_TRUE;
	/**************************/
	int convergenceIterCounter;
	int retCodeTemp;

	ddemivp2ddeComm* ivp2ddeComm;
	ivp2ddeComm = new ddemivp2ddeComm;
	ivp2ddeComm->system = system;
	ivp2ddeComm->tEnd = tEnd;
	ivp2ddeComm->parameters = parameters;
	int nu = system->getNu();
	int omega = system->getOmega();
	int m = system->getNVariables();
	int l = system->getNParameters();

	ivp2ddeComm->alpha = new double[nu+omega];

	ivp2ddeComm->z = new double[m*(nu+omega)];

	ivp2ddeComm->commWithUFuncs = commWithUFuncs;

	ivp2ddeComm->simulator = this;

	ivp2ddeComm->statistics = simulationSolution->statistics;

	ivp2ddeComm->violated = violated;



	parameters_ = parameters;
	system_ = system;
	commWithUFuncs_ = commWithUFuncs;
	simulationSolution_ = simulationSolution;
	

	if(vanishingConst <= 0.0) 
		vanishingConst = 1e-12;
	
	retCodeTemp = prepareSimulationSolution_();
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	/*--------------*/
	double t;
	double* y;
	y = new double[system_->getNVariables()];
	ivp2ddeComm->yTn = y;
	t = simulationSolution_->meshPoints[simulationSolution_->nMeshPoints - 1];
	ivp2ddeComm->addTn = &t;
	for(int i=0; i < system_->getNVariables(); i++)
		y[i] = simulationSolution_->ys[simulationSolution_->nMeshPoints - 1][i];

	violated_ = violated;

	int	vanishingDelay;
	retCodeTemp = 
		updateAlphaAndBaseIndices_(simulationSolution_->meshPoints[simulationSolution_->nMeshPoints - 1]/*double t*/,
																		simulationSolution_->ys[simulationSolution_->nMeshPoints - 1]/*double* y*/
																		 );
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	retCodeTemp =
		findSmallDelays_(simulationSolution_->meshPoints[simulationSolution_->nMeshPoints - 1]/*double t*/,
										violated_->isVanishing, vanishingDelay, vanishingConst);
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	if(vanishingDelay == DDEM_TRUE){
		if(DDEM_TRACE_SIMULATOR){
			cout << "ddemSimulator::simulateWhileContinuous" << endl
							<< "advanced or vanishing delay  at t = " << t << endl;
			for(int i=0; i < l; i++)
				cout << "parameters[" << i << "] = " << parameters[i] << endl;
		}
		ivpCallMode_ = DDEM_VANISHING_CONVERGENCE_LOOP;
	}else
		ivpCallMode_ = DDEM_USUAL_STEPPING;
	
	IVP2DDE_->prepare(system->getNVariables(),
															relTolerance,
															absTolerance);
	IVP2DDE_->restart();

	int IVPError = DDEM_FALSE;
	int discontinuityHappened = DDEM_FALSE;

	int	maxMeshPointsReached;
	if(simulationSolution->nMeshPoints >= DDEM_MAX_SPACE_MESH_POINTS)
				maxMeshPointsReached = DDEM_TRUE;
	else
		maxMeshPointsReached = DDEM_FALSE;
	
	int	maxDiscPointsReached;
	if(simulationSolution->nLambdas >= DDEM_MAX_SPACE_LAMBDAS)
		maxDiscPointsReached = DDEM_TRUE;
	else
		maxDiscPointsReached = DDEM_FALSE;
	
	int ivpReturnCode = 0;
	/*start stepping*/
	if( t < tEnd)
		atEnd = DDEM_FALSE;
	else
		atEnd = DDEM_TRUE;
	convergenceIterCounter = 0;
	while((atEnd == DDEM_FALSE)
				&& (IVPError == DDEM_FALSE)
				&&(discontinuityHappened == DDEM_FALSE)
				&&(maxMeshPointsReached == DDEM_FALSE)
				&&(maxDiscPointsReached == DDEM_FALSE)
				&&(convergenceIterCounter < DDEM_MAX_ITER_PER_VANISH_STEP)){
		violated_->clear();
		ivpReturnCode = IVP2DDE_->tryStep(t, tEnd,
															y,
															simIVPfcn, (void*) ivp2ddeComm);
		if(ivpReturnCode == DDEM_AT_END)
			atEnd = DDEM_TRUE;
		else if(ivpReturnCode == DDEM_ACCEPT){
			//cout << "A";
			double* yTStarByOldInterp;
			yTStarByOldInterp = new double[simulationSolution_->getNVariables()];
			double tStar;
			if(ivpCallMode_ == DDEM_VANISHING_CONVERGENCE_LOOP){
				if(convergenceIterCounter > 0){
			
					tStar = simulationSolution_->meshPoints[simulationSolution_->nMeshPoints - 1];
					for(int i=0; i < m; i++)
						yTStarByOldInterp[i] = simulationSolution_->ys[simulationSolution_->nMeshPoints - 1][i];
					simulationSolution_->nMeshPoints -= 1;
				}
			}
			IVP2DDE_->storeStep(simulationSolution_->meshPoints[simulationSolution_->nMeshPoints],
												simulationSolution_->ys[simulationSolution_->nMeshPoints],
												simulationSolution_->interpolants[simulationSolution_->nMeshPoints-1],
												simulationSolution_->interpStepSizes[simulationSolution_->nMeshPoints-1]);
			simulationSolution_->statistics->nSteps += 1;
			if(ivpCallMode_ == DDEM_VANISHING_CONVERGENCE_LOOP){
				if(convergenceIterCounter > 0){
					/*check for convergence*/
					double* yTStarByNewInterp;
					yTStarByNewInterp = new double[simulationSolution_->getNVariables()];
					IVP2DDE_->interpolate(simulationSolution_->getNVariables(),
																	simulationSolution_->meshPoints[simulationSolution_->nMeshPoints-1],
																	simulationSolution_->ys[simulationSolution_->nMeshPoints-1],
																	tStar,
																	yTStarByNewInterp,
																	simulationSolution_->interpStepSizes[simulationSolution_->nMeshPoints-1],
																	simulationSolution_->interpolants[simulationSolution_->nMeshPoints-1],
																	DDEM_INTERP);
					int errorSatisfied = DDEM_TRUE;
					for(int i=0; i < simulationSolution_->getNVariables(); i++)
						if(fabs(yTStarByNewInterp[i] - yTStarByOldInterp[i]) > 
								yTStarByOldInterp[i] * relTolerance + absTolerance[i]){
							errorSatisfied = DDEM_FALSE;
							break;
						}
					if(errorSatisfied == DDEM_TRUE){
						ivpCallMode_ = DDEM_USUAL_STEPPING;
						simulationSolution_->statistics->nIteraions4VanishingSteps += convergenceIterCounter;
						if(DDEM_TRACE_SIMULATOR)
							cout << endl << "convergenceIterCounter = " << convergenceIterCounter;
					}else
						simulationSolution_->nMeshPoints += 1;
					delete yTStarByNewInterp; yTStarByNewInterp = NULL;
				}
				else{
					simulationSolution_->nMeshPoints += 1;
					simulationSolution_->statistics->nVanishingSteps += 1;
				}
				convergenceIterCounter += 1;
			}
			delete yTStarByOldInterp; yTStarByOldInterp = NULL;
			if(ivpCallMode_ == DDEM_VANISHING_CONVERGENCE_LOOP)
				IVP2DDE_->restart();
			else{
				if(system_->stateDependent())
					violated_->clear();
					computeViolations_(simulationSolution_->meshPoints[simulationSolution_->nMeshPoints],
														 simulationSolution_->ys[simulationSolution_->nMeshPoints],
														 violated_);
				if(violated_->any == DDEM_TRUE){
					//cout << "V";
					int  violType;
					double tHit;
					retCodeTemp = 
						locateTHit_(simulationSolution_->meshPoints[simulationSolution_->nMeshPoints-1],
							simulationSolution_->ys[simulationSolution_->nMeshPoints-1],
							simulationSolution_->interpStepSizes[simulationSolution_->nMeshPoints-1],
							simulationSolution_->interpolants[simulationSolution_->nMeshPoints-1],
							relTolerance,
							tHit,
							violType);
					simulationSolution_->statistics->nRootFindings += 1;
					if(retCodeTemp != DDEM_SUCCESS)
						return retCodeTemp;
					/**/
					int simulatorRequiredContLevel = IVP2DDE_->getRequiredContinuity() + system->getContReductionFactor();
					simulationSolution_->meshPoints[simulationSolution_->nMeshPoints] = tHit;
					retCodeTemp = 
						IVP2DDE_->interpolate(simulationSolution_->getNVariables(),
																	simulationSolution_->meshPoints[simulationSolution_->nMeshPoints-1],
																	simulationSolution_->ys[simulationSolution_->nMeshPoints-1],
																	simulationSolution_->meshPoints[simulationSolution_->nMeshPoints],
																	simulationSolution_->ys[simulationSolution_->nMeshPoints],
																	simulationSolution_->interpStepSizes[simulationSolution_->nMeshPoints-1],
																	simulationSolution_->interpolants[simulationSolution_->nMeshPoints-1],
																	DDEM_INTERP);
					simulationSolution_ ->statistics->nInterpEvals += 1;
					if(retCodeTemp != DDEM_SUCCESS)
						return retCodeTemp;
					if((violType == DDEM_VIOL_DISC)&&
						(violated_->minPropagDiscDegree <= simulatorRequiredContLevel)){
							/* add a new discontinuity point*/
						simulationSolution_->lambdasIndices[simulationSolution_->nLambdas] = simulationSolution_->nMeshPoints;
						simulationSolution_->lambdasDiscDegrees[simulationSolution_->nLambdas] = violated_->minPropagDiscDegree;
						simulationSolution_->nLambdas += 1;
						if(simulationSolution->nLambdas >= DDEM_MAX_SPACE_LAMBDAS)
							maxDiscPointsReached = DDEM_TRUE;
						if(simulationSolution_->lambdasDiscDegrees[simulationSolution_->nLambdas - 1] <= discontinuityLevel)
							discontinuityHappened = DDEM_TRUE;
				}else{
					//cout << "O";
				}
				t = simulationSolution_->meshPoints[simulationSolution_->nMeshPoints];
				for(int j =0; j < system_->getNVariables(); j++)
					y[j] = simulationSolution_->ys[simulationSolution_->nMeshPoints][j]; 
				IVP2DDE_->restart();
				}
				simulationSolution->nMeshPoints += 1;
				if( (tEnd - t) < fabs(t) * relTolerance){
					atEnd = DDEM_TRUE;
					simulationSolution_->meshPoints[simulationSolution_->nMeshPoints-1] = tEnd;
					break;
				}
				double tNew;
				double* yNew;
				tNew = simulationSolution_->meshPoints[simulationSolution_->nMeshPoints - 1];
				yNew = simulationSolution_->ys[simulationSolution_->nMeshPoints - 1];
				//-------------------jfsgsf
				retCodeTemp = updateAlphaAndBaseIndices_(tNew, yNew);
				if(retCodeTemp != DDEM_SUCCESS)
					return retCodeTemp;
				retCodeTemp = 
					findSmallDelays_(tNew, violated_->isVanishing, vanishingDelay, vanishingConst);
				if(vanishingDelay == DDEM_TRUE){
					if(DDEM_TRACE_SIMULATOR){
						cout << "ddemSimulator::simulateWhileContinuous" << endl
										<< "advanced or vanishing delay  at t = " << t << endl;
						for(int i=0; i < l; i++)
							cout << "parameters[" << i << "] = " << parameters[i] << endl;
					}
					ivpCallMode_ = DDEM_VANISHING_CONVERGENCE_LOOP;
					convergenceIterCounter = 0;
				}else
					ivpCallMode_ = DDEM_USUAL_STEPPING;
				if(retCodeTemp != DDEM_SUCCESS)
					return retCodeTemp;
				if(simulationSolution->nMeshPoints >= DDEM_MAX_SPACE_MESH_POINTS)
					maxMeshPointsReached = DDEM_TRUE;
			}
		}else if(ivpReturnCode == DDEM_REJECT){
			//cout << "R";
			simulationSolution->statistics->nFails += 1;
		}else if(ivpReturnCode < 0){
			if(DDEM_TRACE_SIMULATOR)
				cout << "ddemSimulator::simulate"<< endl <<
					"<< IVP integrator returned with error code >> : " 
					<< ivpReturnCode << endl ;
			IVPError = DDEM_TRUE;
		}
	}
	int returnCode;
	if(IVPError == DDEM_TRUE)
		returnCode = ivpReturnCode;
	else if(maxMeshPointsReached == DDEM_TRUE)
		returnCode = DDEM_ERR_SIM_MAX_MESH_POINTS_REACHED;
	else if(maxDiscPointsReached == DDEM_TRUE)
		returnCode = DDEM_ERR_SIM_MAX_DISC_POINTS_REACHED;
	else if(convergenceIterCounter >= DDEM_MAX_ITER_PER_VANISH_STEP)
		returnCode = DDEM_ERR_SIM_VANISHING_ITER_CANNOT_CONVERGE;
	else //if(
		returnCode = DDEM_SUCCESS;
	/**************************/
	simulating_ = DDEM_FALSE;
	/**************************/
	delete y; y = NULL;

	delete ivp2ddeComm->alpha;
	delete ivp2ddeComm->z;

	delete 	ivp2ddeComm;

	/**************************/
	return returnCode;
}
/************************************************************/
int ddemSimulator::simulate(ddemSystemInterface* system, double tEnd,
                  double* parameters,
									double relTolerance,
									double* absTolerance,
									double vanishingConst,
									void* commWithUFuncs,
                  ddemSimulationSolution* simulationSolution)
{
	ddemViolation* violated;
	violated = new ddemViolation(system->getNu() + system->getOmega());
	int atEnd;

	int retCode;
	retCode = simulateWhileContinuous(system,
                          tEnd,
                          parameters,
													relTolerance,
													absTolerance,
													vanishingConst,
                          commWithUFuncs,
                          DDEM_UNDEFINED,/*int discontinuityLevel*/
													violated,
													atEnd,
													simulationSolution);
	delete violated;
	return retCode;
}
/************************************************************/
int	ddemSimulator::computeViolations_(double tRight, double* yRight,
																			ddemViolation*  violated)
{
	
	int retCodeTemp =
		system_->evalDelayArguments(tRight/*double t*/,
																		yRight/*double* y*/,
																		parameters_/*double* p*/,
																		violated->alphasAtTRight/*double* alpha*/,
																		commWithUFuncs_/*void* comm*/);
	simulationSolution_->statistics->nDelayEvals += 1;
	int nu = system_->getNu();
	int omega = system_->getOmega();

	for(int i=0; i < nu + omega; i++)
			evaluateCExtension(tRight, violated->alphasAtTRight[i], NULL,
																			 i,
																			 DDEM_NONE);
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemSimulator::locateTHit_(double tLeftMeshPoint,
													double* yLeftMeshPoint,
													double stepSize,
													double** interpolant,
													double errTol,
													double& tHit, int&  violType)
{
	int retCode;

	double t0, t1, t;

	int nu = system_->getNu();
	int omega = system_->getOmega();

	t0 = tLeftMeshPoint;
	double tRightMin;
	double* g0,* g1,* gt;
	int* active;
	int nfcnCall;
	int ng;
	
	ddemDelays2MinRootComm* comm;
	comm = new ddemDelays2MinRootComm;
	comm->candidateActiveIndices = new int[nu + omega];
	/*--*/
	comm->y = new double[system_->getNVariables()];/* no initialization required*/
	comm->alpha = new double[nu + omega];/* no initialization required*/

	comm->nVariables = system_->getNVariables();
	comm->tLeftMeshPoint = tLeftMeshPoint;
	comm->yLeftMeshPoint = yLeftMeshPoint;
	comm->stepSize = stepSize;
	comm->interpolant = interpolant;
	comm->parameters = parameters_;
	comm->commWithUFuncs = commWithUFuncs_;

	comm->tZero = violated_->tZero;

	comm->system = system_;
	comm->IVP2DDE = IVP2DDE_;

	comm->statistics = simulationSolution_->statistics;
	/*--*/
	g0 = new double[nu + omega];
	int firstTime = DDEM_TRUE;
	ng = 0;
	for(int j=0; j < nu + omega; j++)
		if(violated_->before[j] == DDEM_TRUE){
			g0[ng] = violated_->alphasAtTn[j] - violated_->tZero[j];
			comm->candidateActiveIndices[ng] = j;
			ng+=1;
			if(firstTime){
				tRightMin = violated_->tSource[j];
				firstTime = DDEM_FALSE;
			}else if (violated_->tSource[j] < tRightMin)
				tRightMin = violated_->tSource[j];
		}
	t1 = tRightMin; 
	g1 = new double[ng];
	gt = new double[ng];
	active = new int[ng];
	/*--*/
	comm->ng = ng;
	/*--*/
	ddemDelay2g(t1, g1, comm);
	int errCodeRoot = ddemMinRoot(t0, t1, ng, g0, g1, ddemDelay2g,
														errTol, t, gt, active, (void*)comm, nfcnCall);
	if(errCodeRoot == DDEM_SUCCESS){
		if(DDEM_TRACE_SIMULATOR)
		{
			cout << "Minimum root = " << t << endl
					<<"Number of function evaluations = " << nfcnCall << endl;
			for(int i=0; i < ng; i++)
				cout << "active[" << i <<"] = " << active[i]<< endl;
		}
		/*--------------------------*/
		tHit = t;
		violType = DDEM_VIOL_ADVANCED;
		for(int k=0; k < ng ; k++)
			if((active[k] == 1) &&
				(violated_->violType[comm->candidateActiveIndices[k]] ==
				DDEM_VIOL_DISC)){
				violType = DDEM_VIOL_DISC;
				violated_->violAlphaIndex = comm->candidateActiveIndices[k];
				break;
			}
		int minPropagDiscDegree = 1000;
		for(int k=0; k < ng ; k++)
			if((active[k] == 1) &&
				(violated_->violType[comm->candidateActiveIndices[k]] ==
				DDEM_VIOL_DISC)){
				int violAlphaIndex = comm->candidateActiveIndices[k];
				int discIndex;
				if(violated_->direction[violAlphaIndex] == DDEM_PLUS)
					discIndex = violated_->baseIndices[violAlphaIndex] + 1;
				else
					discIndex = violated_->baseIndices[violAlphaIndex] - 1;
				int discDegree;
				if(discIndex >= 0){
					discDegree = simulationSolution_->lambdasDiscDegrees[discIndex];
				}else{
					discDegree = simulationSolution_->historyStartsDiscDegrees[-discIndex - 1];
				}
				int nu = system_->getNu();
				int propagDiscDegree;
				propagDiscDegree = discDegree + system_->getContPropagFactor(violAlphaIndex);
				if(propagDiscDegree == 0)
					propagDiscDegree = 1;
				minPropagDiscDegree = ddem_min(minPropagDiscDegree, propagDiscDegree);
			}
		violated_->minPropagDiscDegree = minPropagDiscDegree;
		retCode = DDEM_SUCCESS;
	}else{
		if(DDEM_TRACE_SIMULATOR)
			cout << "Failure  in ddemMinRoot" << endl 
			     <<"Error Code = " << errCodeRoot << endl;
		retCode = errCodeRoot;
	}
	delete comm->candidateActiveIndices;
	delete comm->y;
	delete comm->alpha;
	delete comm;

	delete g0;
	delete g1;
	delete gt;
	delete active;

	return retCode;
}
/************************************************************/
int ddemSimulator::prepareSimulationSolution_()
{
	if(simulationSolution_->isAllocated() == DDEM_FALSE){
		int nHistorySegments = system_->getNHistorySegments();
		int retCode = 
			simulationSolution_->allocateMemories(system_->getNVariables(),
																	IVP2DDE_->getNInterploatorCoefficens(),
																	nHistorySegments);
		if(retCode != DDEM_SUCCESS)
			return retCode;

		int retCodeTemp;
		
		for(int i=0; i < nHistorySegments - 1; i++){
			retCodeTemp = system_->evalHistoryStart(i + 1,
																parameters_,
																simulationSolution_->historyStarts[i],
																commWithUFuncs_);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
			retCodeTemp = system_->getHistoryStartDiscDegree(i + 1,
																simulationSolution_->historyStartsDiscDegrees[i],
																commWithUFuncs_);
			if(retCodeTemp != DDEM_SUCCESS)
				return retCodeTemp;
		}
		//}
	}
	if(simulationSolution_->nMeshPoints == 0){
		int retCode = 
			system_->compT0(parameters_, simulationSolution_->meshPoints[0],
										commWithUFuncs_);
		if(retCode != DDEM_SUCCESS)
			return retCode;
		retCode = 
			system_->compY0(parameters_, simulationSolution_->ys[0],
										simulationSolution_->historyStarts,
									commWithUFuncs_);
		if(retCode != DDEM_SUCCESS)
			return retCode;
		simulationSolution_->nMeshPoints = 1;
		simulationSolution_->lambdasIndices[0] = 0;
		system_->compT0DiscDegree(parameters_, simulationSolution_->lambdasDiscDegrees[0],
															commWithUFuncs_);
		simulationSolution_->nLambdas = 1;
	}
	return DDEM_SUCCESS;
}
/************************************************************/
ddemSimulator::ddemSimulator(ddemIVP2DDE* ivp2dde)
{
	simulating_ = DDEM_FALSE;
	IVP2DDE_ = ivp2dde;
}
/************************************************************/
ddemSimulator::~ddemSimulator()
{
}
/************************************************************/
int ddemSimulator::updateAlphaAndBaseIndices_(double t, double* y)//, int& vanishingDelay)
{
	int retCodeTemp =
		system_->evalDelayArguments(t/*double t*/,
																		y/*double* y*/,
																		parameters_/*double* p*/,
																		violated_->alphasAtTn/*double* alpha*/,
																		commWithUFuncs_/*void* comm*/);
	simulationSolution_->statistics->nDelayEvals += 1;
	if(retCodeTemp != DDEM_SUCCESS)
		return retCodeTemp;
	int nu = system_->getNu();
	int omega = system_->getOmega();
	for(int i=0; i < nu + omega ; i++){
		int retCodeLocate =
			locateBetweenLambdasOrHistorySegments_(violated_->alphasAtTn[i], violated_->baseIndices[i]);
		if(retCodeLocate != DDEM_SUCCESS)
			return retCodeLocate;
	}
	return DDEM_SUCCESS;
}
int ddemSimulator::findSmallDelays_(double t, int* isVanishing, int& vanishingDelay, double vanishingConst)
{
	int nu = system_->getNu();
	int omega = system_->getOmega();
	vanishingDelay = DDEM_FALSE;
	for(int i=0; i < nu + omega ; i++){
		if((violated_->alphasAtTn[i] >= t - vanishingConst))
		{
			isVanishing[i] = DDEM_TRUE;
			vanishingDelay = DDEM_TRUE;
		}else
			isVanishing[i] = DDEM_FALSE;

	}
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemSimulator::evaluateCExtension(double tSource, double t, double* y,
																			 int iBaseIndex,
																			 int interpSelect)
{
	if(!simulating_){
		if(DDEM_TRACE_SIMULATOR)
			cout << "ddemSimulator::evaluateCExtension" << " " <<
			"can only be called during the simulation!" << endl;
		return DDEM_ERR_SIM_CALL_DENIED;
	}
	int baseIndex = violated_->baseIndices[iBaseIndex];
	if(baseIndex >= 0){/* in the solution area*/
		int CEIndex;
		if(t < simulationSolution_->meshPoints
			[simulationSolution_->lambdasIndices[baseIndex]]){
				if(simulationSolution_->nMeshPoints > 1){
					CEIndex = simulationSolution_->lambdasIndices[baseIndex];
					violation_(DDEM_VIOL_DISC, DDEM_MINUS, iBaseIndex, tSource, t,
						simulationSolution_->meshPoints[CEIndex]);
				}else
					return DDEM_ERR_SIM_REQUIRED_EXTENSION_IS_IMPOSSIBLE;
		}else if(baseIndex == simulationSolution_->nLambdas - 1){
						if(simulationSolution_->lambdasIndices[baseIndex] ==
													simulationSolution_->nMeshPoints - 1)
								/* base discontinuity point is also the last mesh point */
								/* and cannot be a base for extrapolation*/
								return DDEM_ERR_SIM_REQUIRED_EXTENSION_IS_IMPOSSIBLE;
						else if(t >= simulationSolution_->meshPoints
							[simulationSolution_->nMeshPoints - 1]){
							CEIndex = simulationSolution_->nMeshPoints - 2;
							violation_(DDEM_VIOL_ADVANCED, DDEM_PLUS, iBaseIndex, tSource, t,
							simulationSolution_->meshPoints[simulationSolution_->nMeshPoints - 1]);
						}else{
							int retCodeLocate2;
							retCodeLocate2 = locateBetweenMeshPointsSolution(t, baseIndex, CEIndex, simulationSolution_);
							if(retCodeLocate2 != DDEM_SUCCESS)
								return retCodeLocate2;
						}
		}else if(t >= simulationSolution_->meshPoints
			[simulationSolution_->lambdasIndices[baseIndex + 1]]){
					CEIndex = simulationSolution_->lambdasIndices[baseIndex + 1] - 1;
					violation_(DDEM_VIOL_DISC, DDEM_PLUS, iBaseIndex, tSource, t,
					simulationSolution_->meshPoints[CEIndex + 1]);
		}else{ /* in between*/
			int retCodeLocate2;
			retCodeLocate2 = locateBetweenMeshPointsSolution(t, baseIndex, CEIndex, simulationSolution_);
			if(retCodeLocate2 != DDEM_SUCCESS)
				return retCodeLocate2;
		}
		if(interpSelect != DDEM_NONE){
			simulationSolution_->statistics->nInterpEvals += 1;
			return evaluateSolution2(t, y, simulationSolution_, interpSelect, CEIndex);
		}
	}else{/* in the history*/
		if(-baseIndex <  simulationSolution_->getNHistorySegments()){
			double tLeft;
			tLeft = simulationSolution_->historyStarts[-baseIndex - 1];
			if(t < tLeft)
				violation_(DDEM_VIOL_DISC, DDEM_MINUS, iBaseIndex, tSource, t, tLeft);
		}
		double tRight;
		if(-baseIndex == 1){
			tRight = simulationSolution_->meshPoints[0];
			if(t >= tRight)
				violation_(DDEM_VIOL_DISC, DDEM_PLUS, iBaseIndex, tSource, t, tRight);
		}else{
			tRight = simulationSolution_->historyStarts[-baseIndex - 2];
			if(t >= tRight)
				violation_(DDEM_VIOL_DISC, DDEM_PLUS, iBaseIndex, tSource, t, tRight);
		}
		if(interpSelect == DDEM_INTERP){
			simulationSolution_->statistics->nHistoryEvals += 1;
			return system_->evalHistory(-baseIndex, t, parameters_, y,
																	commWithUFuncs_);
		}else if(interpSelect == DDEM_DERIV_INTERP){
			simulationSolution_->statistics->nHistoryEvals += 1;
			return system_->evalDerivHistory(-baseIndex, t, parameters_, y,
																	commWithUFuncs_);
		}
	}
	return DDEM_SUCCESS;
}
/************************************************************/
void ddemSimulator::violation_(int violType, int direction, int iBaseIndex,
										double tSource, double t, double tZero)
{
	if(DDEM_TRACE_SIMULATOR)
	{
		cout << "violation at " << tSource << " Type =  " ;
		if(violType == DDEM_VIOL_ADVANCED)
			cout << "DDEM_VIOL_ADVANCED";
		else
			cout << "DDEM_VIOL_DISC";
		cout << "  direction = ";
		if(direction == DDEM_PLUS)
			cout << "+";
		else
			cout << "-";
		cout << endl;
	}
	if((violated_->isVanishing[iBaseIndex] == DDEM_TRUE) && 
		 (violType == DDEM_VIOL_ADVANCED))
		return;
	violated_->any = DDEM_TRUE;
	if(violated_->before[iBaseIndex] == DDEM_TRUE){
		if(tSource < violated_->tSource[iBaseIndex]){
			violated_->tSource[iBaseIndex] = tSource;
			violated_->t[iBaseIndex] = t;
			violated_->tZero[iBaseIndex] = tZero;
			violated_->violType[iBaseIndex] = violType;
			violated_->direction[iBaseIndex] = direction;
		}
	}else{
		violated_->before[iBaseIndex] = DDEM_TRUE;
		violated_->tSource[iBaseIndex] = tSource;
		violated_->t[iBaseIndex] = t;
		violated_->tZero[iBaseIndex] = tZero;
		violated_->violType[iBaseIndex] = violType;
		violated_->direction[iBaseIndex] = direction;
	}
}
/************************************************************/
int ddemSimulator::evaluateSolution2(double t, double* y,
														 ddemSimulationSolution* simulationSolution,
														 int interpSelect,
														 int betweenMeshPointsIndex)
{
	if((betweenMeshPointsIndex >= 0)&&
		(betweenMeshPointsIndex <= simulationSolution->nMeshPoints - 1)){
		if(betweenMeshPointsIndex == simulationSolution->nMeshPoints - 1)
			betweenMeshPointsIndex = simulationSolution->nMeshPoints - 2;
		return IVP2DDE_->interpolate(simulationSolution->getNVariables(),
														simulationSolution->meshPoints[betweenMeshPointsIndex],
														simulationSolution->ys[betweenMeshPointsIndex],
														t, y,
														simulationSolution->interpStepSizes[betweenMeshPointsIndex],
														simulationSolution->interpolants[betweenMeshPointsIndex],
														interpSelect);
	}else{
		if(DDEM_TRACE_SIMULATOR)
			cout << "(ddemSimulator::evaluateSolution2) mesh point index out of range " << endl;
		return DDEM_ERR_SIM_MESH_POINT_INDEX_OUT_OF_RANGE;
	}
}
/************************************************************/
int ddemSimulator::evaluateSolution(double t, double* y,
														 ddemSimulationSolution* simulationSolution,
														 int interpSelect,
														 int betweenLambdasIndex)
{
	int retCodeLocate1;
	if(betweenLambdasIndex < 0)/* not provided by caller or out of range*/
		retCodeLocate1 = locateBetweenLambdasSolution(t, betweenLambdasIndex, simulationSolution);
	else if(betweenLambdasIndex < simulationSolution ->nLambdas)
		retCodeLocate1 = DDEM_SUCCESS;
	else{
		if(DDEM_TRACE_SIMULATOR)
			cout << "(ddemSimulator::evaluateSolution) lambda index out of range " << endl;
		retCodeLocate1 = DDEM_ERR_SIM_LAMBDA_INDEX_OUT_OF_RANGE;
	}
	if(retCodeLocate1 == DDEM_SUCCESS){
		int betweenMeshPointsIndex;
		int retCodeLocate2;
		retCodeLocate2 = locateBetweenMeshPointsSolution(t, betweenLambdasIndex, betweenMeshPointsIndex,
																				simulationSolution);
		if( retCodeLocate2 == DDEM_SUCCESS)
			return evaluateSolution2(t, y,
														 simulationSolution,
														 interpSelect,
														 betweenMeshPointsIndex);
		else
			return retCodeLocate2;
	}else
		return retCodeLocate1;
}
/************************************************************/
int ddemSimulator::locateBetweenMeshPointsSolution(double t, int betweenLambdasIndex,
																						int& betweenMeshPointsIndex,
																						ddemSimulationSolution* simulationSolution)
{
	int leftMeshIndex = simulationSolution->lambdasIndices[betweenLambdasIndex];
	int rightMeshIndex; 
	if(betweenLambdasIndex == simulationSolution->nLambdas-1){
		if(t > simulationSolution->meshPoints[simulationSolution->nMeshPoints - 1]){
			if(DDEM_TRACE_SIMULATOR)
			{
				cout << "<<    ddemSimulator::locateBetweenMeshPointsSolution     >>" << endl;
				cout << " Advanced point" << endl;
			}
			return DDEM_ERR_SIM_T_GREATER_THAN_TLAST;
		}else
			rightMeshIndex = simulationSolution->nMeshPoints-1;
	}else
		rightMeshIndex = simulationSolution->lambdasIndices[betweenLambdasIndex + 1]; 
	while (leftMeshIndex + 1  < rightMeshIndex){/*binary search*/
		int midMeshIndex = (leftMeshIndex + rightMeshIndex)/2;
		if(simulationSolution->meshPoints[midMeshIndex] <= t)
			leftMeshIndex = midMeshIndex;
		else
			rightMeshIndex = midMeshIndex;
	}
	betweenMeshPointsIndex = leftMeshIndex;
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemSimulator::locateBetweenLambdasSolution(double t, int& betweenLambdasIndex,
																								ddemSimulationSolution* simulationSolution)
{
	if(t < simulationSolution->meshPoints[0]) 
		return DDEM_ERR_SIM_T_LESS_THAN_T0;
	int candidateIndex = simulationSolution->nLambdas - 1;
	/* search the solution after history */
	/* return 0 1 2 ... (nLambdas - 1) */
	while(candidateIndex >= 0){
		if(simulationSolution->
			meshPoints[simulationSolution->lambdasIndices[candidateIndex]] 
			<= t ){
			betweenLambdasIndex = candidateIndex;
			return DDEM_SUCCESS;
		}
		else
			candidateIndex = candidateIndex - 1;
	}
	if(DDEM_TRACE_SIMULATOR)
		cout <<"<<Unexpected situation in ddemSimulator::locateBetweenLambdasSolution >>";
	return DDEM_ERR_UNEXPECTED;
}
/************************************************************/
int ddemSimulator::locateBetweenLambdasOrHistorySegments_(double t, int& betweenLambdasIndex)
{
	int retCode;
	if(t >= simulationSolution_->meshPoints[0])
		retCode = locateBetweenLambdasSolution(t, betweenLambdasIndex, simulationSolution_);
	else{
		/* search the history */
		/* betweenLambdasIndex = -1 -2 -3 ... -nHistorySegments*/
		retCode = system_->locateBetweenHistoryStarts(t, simulationSolution_->historyStarts,
																					betweenLambdasIndex);
		betweenLambdasIndex = - betweenLambdasIndex;
	}
	return retCode;
}
/************************************************************/
