/* DOUBLE CHECKED = YES*/
#include <iostream>
using namespace std;
#include <fstream>
#include <iomanip>
#include <string>

#include "ddem_sol_out.h"
#include "ddem_error_codes.h"
#include "ddem_trace.h"

/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
std::string ddemInt2string(int i)
{
	const char* conversionTable[] = {"00","01","02","03","04","05","06","07","08","09","10","11","12",
													"13","14","15","16","17","18","19"};
	return std::string(conversionTable[i]);
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
enum {DDEM_OUT_SOL_SINGLE, DDEM_OUT_SOL_ALL};
int ddemOutSolUniformBase_(ddemSimulationSolution*	simulationSolution,
											int outSolType,
											int outYIndex,
											int interpSelect,
											int nOutPoints,
											string fileName,
											int precision,
											ddemSimulator* simulator)
{
	ofstream outSolFile;
	outSolFile.open(fileName.c_str(), ios::out );
	if (!outSolFile){
		cerr << "File " << fileName << " could not be opened" << endl;
		return DDEM_FAILURE;
	}

	double tStart = simulationSolution->meshPoints[0];
	double tEnd = simulationSolution->meshPoints[simulationSolution->nMeshPoints - 1];
	double* y;
	y = new double[simulationSolution->getNVariables()];

	outSolFile.unsetf(ios::fixed);
	outSolFile <<  setiosflags( ios::scientific | ios::showpoint ) << setprecision(precision);
	ddemUniformStepper* uniformStepper; 
	uniformStepper = new ddemUniformStepper();
	uniformStepper->initialize(tStart, tEnd, nOutPoints);

	double t;
	for(int i=0; i < nOutPoints ; i++){
		if(uniformStepper->get(i,t) == DDEM_SUCCESS){
			simulator->evaluateSolution(t, y, simulationSolution, DDEM_INTERP, DDEM_UNKNOWN);
			outSolFile << t;
			if(outSolType == DDEM_OUT_SOL_SINGLE)
							outSolFile << "  " << y[outYIndex];
			else
				for(int q = 0; q < simulationSolution->getNVariables(); q++)
					outSolFile << "  " << y[q];
			outSolFile << endl;
		}
	}
	delete uniformStepper;
	delete y;
	outSolFile.close();
	return DDEM_SUCCESS;
}
/************************************************************/
/************************************************************/
int ddemOutSolUniform(ddemSimulationSolution*	simulationSolution,
											int outYIndex,
											int interpSelect,
											int nOutPoints,
											string fileName,
											int precision,
											ddemSimulator* simulator)
{
	return ddemOutSolUniformBase_(simulationSolution,
											DDEM_OUT_SOL_SINGLE,
											outYIndex,
											interpSelect,
											nOutPoints,
											fileName,
											precision,
											simulator);
}
/************************************************************/
/************************************************************/
int ddemOutSolUniformAll(ddemSimulationSolution*	simulationSolution,
											int interpSelect,
											int nOutPoints,
											string fileName,
											int precision,
											ddemSimulator* simulator)
{
	return ddemOutSolUniformBase_(simulationSolution,
											DDEM_OUT_SOL_ALL,
											DDEM_UNDEFINED,
											interpSelect,
											nOutPoints,
											fileName,
											precision,
											simulator);
}
/************************************************************/
/************************************************************/
int ddemOutSolNonUniformBase_(ddemSimulationSolution*	simulationSolution,
											int outSolType,
											int outYIndex,
											int interpSelect,
											int nOutPoints,
											string fileName,
											string fileNameDisc,
											int precision,
											ddemSimulator* simulator)
{
	int nOutPointsInEachInterval = ddem_max(3, nOutPoints
																						/ (simulationSolution ->nMeshPoints - 1));
	ofstream outSolFile;
	outSolFile.open( fileName.c_str(), ios::out );
	if (!outSolFile){
		cerr << "File " << fileName << " could not be opened" << endl;
		return DDEM_FAILURE;
	}
	ofstream outSolFileDisc;
	outSolFileDisc.open( fileNameDisc.c_str(), ios::out );
	if (!outSolFileDisc){
		cerr << "File " << fileNameDisc << " could not be opened" << endl;
		return DDEM_FAILURE;
	}
	double* y;
	y = new double[simulationSolution->getNVariables()];

	outSolFile <<  setiosflags( ios::scientific | ios::showpoint ) << setprecision(precision); 
	int currentOutCounter = 1;
	for(int i=0; i < simulationSolution->nLambdas; i++){
		outSolFileDisc <<  currentOutCounter << endl;
		int segmentLeftIndex = simulationSolution->lambdasIndices[i];
		int segmentRightIndex;
		if( i < simulationSolution->nLambdas - 1)
			segmentRightIndex = simulationSolution->lambdasIndices[i + 1];
		else
			segmentRightIndex = simulationSolution->nMeshPoints - 1;
		for(int j = segmentLeftIndex; j < segmentRightIndex; j++){
			double intervalLeft =  simulationSolution->meshPoints[j];
			double intervalRight =  simulationSolution->meshPoints[j + 1];
			double hJ;
			if(nOutPointsInEachInterval > 1)
				hJ = (intervalRight - intervalLeft) / (nOutPointsInEachInterval - 1);
			else{/* nOutPointsInEachInterval == 1 */
				if (j == segmentRightIndex - 1){
					nOutPointsInEachInterval = 2;
					hJ = intervalRight - intervalLeft;
				}else
					hJ = 0.0; /* does not matter */
			}
			int nK;
			if( j == segmentRightIndex - 1)
				nK =  nOutPointsInEachInterval;
			else
				nK =  nOutPointsInEachInterval - 1;
			for(int k = 0; k < nK; k++){
				double t = intervalLeft + k * hJ;
				simulator->evaluateSolution2(t, y, simulationSolution, interpSelect, j);
				outSolFile << t;
				if(outSolType == DDEM_OUT_SOL_SINGLE)
					outSolFile << "  " << y[outYIndex];
				else
					for(int q = 0; q < simulationSolution->getNVariables(); q++)
						outSolFile << "  " << y[q];
				outSolFile << endl;
				currentOutCounter += 1;
			}
		}
	}
	delete y;
	outSolFile.close();
	outSolFileDisc.close();
	return DDEM_SUCCESS;
}
/************************************************************/
/************************************************************/
int ddemOutSolNonUniform(ddemSimulationSolution*	simulationSolution,
											int outYIndex,
											int interpSelect,
											int nOutPoints,
											string fileName,
											string fileNameDisc,
											int precision,
											ddemSimulator* simulator)
{
	return ddemOutSolNonUniformBase_(simulationSolution,
											DDEM_OUT_SOL_SINGLE,
											outYIndex,
											interpSelect,
											nOutPoints,
											fileName,
											fileNameDisc,
											precision,
											simulator);
}
/************************************************************/
/************************************************************/
int ddemOutSolNonUniformAll(ddemSimulationSolution*	simulationSolution,
											int interpSelect,
											int nOutPoints,
											string fileName,
											string fileNameDisc,
											int precision,
											ddemSimulator* simulator)
{
	return ddemOutSolNonUniformBase_(simulationSolution,
											DDEM_OUT_SOL_ALL,
											DDEM_UNDEFINED,
											interpSelect,
											nOutPoints,
											fileName,
											fileNameDisc,
											precision,
											simulator);
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
ddemUniformStepper::ddemUniformStepper()
{
	isInitialized_ = DDEM_FALSE;
}
/************************************************************/
ddemUniformStepper::~ddemUniformStepper()
{
}
/************************************************************/
int ddemUniformStepper::initialize(double tStart, double tEnd, int nPoints)
{
	h_ = (tEnd - tStart) / (nPoints - 1);
	tStart_ = tStart;
	tEnd_ = tEnd;
	nPoints_ = nPoints;
	isInitialized_ = DDEM_TRUE;
	return DDEM_SUCCESS;
}
/************************************************************/
int ddemUniformStepper::get(int index, double& t)
{
	if(isInitialized_ == DDEM_TRUE){
		if((index < 0) || (index >= nPoints_))
			return DDEM_FAILURE;
		else{
			if(index == nPoints_ - 1)
				t = tEnd_; 
			else
				t = tStart_ + index * h_;
			return DDEM_SUCCESS;
		}
	}else{
		cout << "demUniformStepper::next can't be called before ddemUniformStepper::initialize"
				<< endl;
		return DDEM_FAILURE;
	}
}
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
