#ifndef __DDEM_SYSTEM
#define __DDEM_SYSTEM

#include <cstddef>

#include "ddem_basic_defins.h"
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
struct ddemUserProvidedJac{
	int  needed;
	int  needed2nd;

  int tIndependent;
  int pIndependent;
  int yIndependent;
  int zIndependent;
	
	double* tJac;/* 1 Dimensional */
	double* pJac;/* 2 Dimensional */
	double* yJac;/* 2 Dimensional */
	double* zJac;/* 3 Dimensional F_m*(alpha_(nu+omega)*y_m)*/

  int ttIndependent;
  int tpIndependent;

	double* ttJac;
	double* tpJac;

	ddemUserProvidedJac();
	~ddemUserProvidedJac();
	void reset();
};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
typedef void (*ddemF)(int state, double t, int m, double* y,
										  int nu, int omega, double* z,
											int l, double* p,
											double* f,
                      ddemUserProvidedJac* userProvidedJac,
											void* comm);
/* z : (nu+omega) by m */
/* z(1:nu, 1:m) correspond to y_{1:m}(alpha_{1:nu}) */
/* z(nu+1:omega, 1:m) correspond to y'_{1:m}(alpha_{nu+1:nu+omega}) */
/* y_1_alpha_1       y_2_alpha_1       ...  y_m_alpha_1       */
/* y_1_alpha_2       y_2_alpha_2       ...  y_m_alpha_2       */
/* ...                                                        */
/* y_1_alpha_nu      y_2_alpha_nu      ...  y_m_alpha_nu      */
/* y'_1_alpha_(nu+1) y'_2_alpha_(nu+1) ...  y'_m_alpha_(nu+1) */
/* y'_1_alpha_(nu+2) y'_2_alpha_(nu+2) ...  y'_m_alpha_(nu+2) */
/* ...                                                        */
/* y'_1_alpha_(nu+o) y'_2_alpha_(nu+o) ...  y'_m_alpha_(nu+o) *//*o:omega*/

/* output: f */
typedef void (*ddemEvents)(double t, int m, double* y, double* yp,
                                   int l, double* p,
                                   int ng, double* g,
                                   ddemUserProvidedJac* userProvidedJac,
																	 void* comm);

typedef void (*ddemChangeFunc)(int currentState,
																	 int nOccuredEvents,
																	 int* occuredEventsIndices,
																	 int* occuredEventsDirections,
																	 double t,
																	 int m, double* y, double* yp,
																	 int l, double* p,
																	 int& yJump,
																	 double* yNew,
																	 int& newState,
																	 int& terminate);/*{DDEM_TRUE, DDEM_FALSE}*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
class ddemAutomata{
	private:
		int created_;
	protected:
    ddemF fFuncs_;/**/
    ddemEvents events_;/**/
		int* terminal_;
    ddemChangeFunc changeFunc_;/**/
		int nEvents_;
		/*int startState_;*/
		int currentState_;
  public:
		int getCurrentState();
		int getNEvents();
		int create(int nEvents);
		ddemAutomata();
   ~ddemAutomata();
	 	int setF(ddemF fFuncs);
	 	int setStartState(int startState);
	 	int setEvents(ddemEvents events, int* terminal=NULL);
	 	int setChangeFunc(ddemChangeFunc changeFunc);
};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
typedef void (*ddemT0Func)(int l, double* p,
                       double& tStart,
                       ddemUserProvidedJac* userProvidedJac,
											 void* comm);
/* output: tStart */
typedef  ddemT0Func ddemTEndFunc;
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
typedef void (*ddemHistoryStartsFunc)(int historySegmentIndex,
																 int l, double* p,
                       double& historyStart,
                       ddemUserProvidedJac* userProvidedJac,
											 void* comm);
/* output: historyStart */
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
typedef void (*ddemHistoryFunc)(int historySegmentIndex,
																	 double t,
                                   int l, double* p,
                                   int m, double* h,
                                   ddemUserProvidedJac* userProvidedJac,
																	 void* comm);
/* output: h */
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
enum {DDEM_NOT_SPECIFIED, DDEM_SPECIFIED_BY_CONSTANT,
			DDEM_SPECIFIED_BY_FUNC, DDEM_SPECIFIED_BY_HISTORY};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
class ddemHistory{
	private:
		int created_;
  protected:
    int nHistorySegments_;        /*number of segments*/

		double* historyConstant_;/*constant history (nHistorySegments_ = 1)*/
		ddemHistoryFunc historyFunc_;/*functions defining  history segments*/
    int   historyType_;/* */

		double* historyStartsConstant_;    /*Start points of  segments*/
    ddemHistoryStartsFunc historyStartsFunc_;
    int   historyStartsType_;/* */
		
		int* historyStartsDiscDegreesConstant_;

		virtual int getNVariables()=0;
	public:
    int create(int nHistorySegments);
    ddemHistory();
   ~ddemHistory();
    int setHistory(double* historyConstant);
    int setHistory(ddemHistoryFunc historyFunc);
		int setHistoryStarts(double* historyStartsConstant);
		int setHistoryStarts(ddemHistoryStartsFunc historyStartsFunc);
		int getHistoryStartsType(){return historyStartsType_;};

		int setHistoryStartsDiscDegrees(int* historyStartsDiscDegreesConstant);
};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/************************************************************/
typedef void (*ddemY0Func)(int l, double* p,
                       int m, double* y0,
                       ddemUserProvidedJac* userProvidedJac,
											 void* comm);
/* output: y0 */
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
class ddemStartingConditions{
	private:
		int created_;
  protected:
		ddemT0Func t0Func_;
		double t0Constant_;
		int t0Type_;

		int t0DiscDegreeConstant_;
		
		ddemY0Func y0Func_;
		double* y0Constant_;
		int y0Type_;
    virtual int getNVariables()=0;
	public:
		int create();
		ddemStartingConditions();
	 ~ddemStartingConditions();
    int setT0(double t0Constant);
	  int setT0(ddemT0Func t0Func);

		int setT0DiscDegree(int t0DiscDegreeConstant);

    int setY0(double* y0Constant);
	  int setY0(ddemY0Func y0Func);

		int getT0Type(){return t0Type_;};
};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
typedef void (*ddemDelayArguments)(double t, int m, double* y,
                                   int l, double* p,
                                   int nu, int omega, double* alpha,
                                   ddemUserProvidedJac* userProvidedJac,
																	 void* comm);
/* output: alpha */
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
class ddemSystemInterface{
  public:
    virtual int getNVariables()=0;
    virtual int getNParameters()=0;
		virtual int getNHistorySegments()=0;
		virtual int getNu()=0;
		virtual int getOmega()=0;
    virtual int evalF(double t, double* y,
										 double* z,
										 double* p,
										 double* f,
										 void* comm)=0;
		virtual int evalDelayArguments(double t, double* y,
																		double* p,
																		double* alpha,
																		void* comm)=0;
		virtual int evalHistoryStart(int historySegmentIndex,
																	double*p,
																	double& historyStart,
																	void* comm)=0;
		virtual int getHistoryStartDiscDegree(int historySegmentIndex,
																	int& historyStartDiscDegree,
																	void* comm)=0;
    virtual int evalHistory(int historySegmentIndex,
																		double t,
                                    double* p,
                                    double* h,
																	  void* comm)=0;
    virtual int evalDerivHistory(int historySegmentIndex,
																		double t,
                                    double* p,
                                    double* h,
																	  void* comm)=0;
		virtual int compT0(double* p,
               double& t0,
							void* comm)=0;
		virtual int compT0DiscDegree(double* p,
               int& t0DiscDegree,
							void* comm)=0;
		virtual int compY0(double* p,
               double* y0,
							 double* historyStarts,
							 void* comm)=0;
		virtual int locateBetweenHistoryStarts(double t,
																						double* historyStarts,
																						int& betweenHisroryIndex)=0;
		virtual int getContReductionFactor()=0;
		virtual int getContPropagFactor(int delayArgumentIndex)=0;
		virtual int stateDependent()=0;

private:
		virtual void createUserProvidedJacs_()=0;
		virtual void destroyUserProvidedJacs_()=0;
protected:
	ddemUserProvidedJac* userProvidedJac4F_;
	ddemUserProvidedJac* userProvidedJac4HistoryStart_;
	ddemUserProvidedJac* userProvidedJac4History_;
	ddemUserProvidedJac* userProvidedJac4DelayArguments_;
	ddemUserProvidedJac* userProvidedJac4T0_;
	ddemUserProvidedJac* userProvidedJac4Y0_;
};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
class ddemSystem: public ddemAutomata,
                   public ddemHistory,
                   public ddemStartingConditions,
									 public ddemSystemInterface{
	private:
		int created_;
		double infinityBound_;
	private:
		int nParameters_;
		int	nVariables_;
			
    int nu_;
    int omega_;
    ddemDelayArguments delayArgumentsFunc_;
    double* delayArgumentsConstant_;
    int delayArgumentsType_;
		int stateDependent_;
  public:
    virtual int getNVariables();
    virtual int getNParameters();
		virtual int getNHistorySegments();
		virtual int getNu();
		virtual int getOmega();
    int create(int nVariables, int nParameters,
								int nHistorySegments,
								int nEvents);
    ddemSystem();
    ~ddemSystem();
    int setDelayArguments(int nu, int omega, double* delayArgumentsConstant);
    int setDelayArguments(int nu, int omega,
                           ddemDelayArguments delayArgumentsFunc,
													 int stateDependent);
		int checkIntegrity(void);
    virtual int evalF(double t, double* y,
										 double* z,
										 double* p,
										 double* f,
										 void* comm);
    int evalFBase(double t, double* y,
										 double* z,
										 double* p,
										 double* f,
                     ddemUserProvidedJac* userProvidedJac,
										 void* comm);
		virtual int evalDelayArguments(double t, double* y,
																		double* p,
																		double* alpha,
																		void* comm);
		int evalDelayArgumentsBase(double t, double* y,
																		double* p,
																		double* alpha,
																		ddemUserProvidedJac* userProvidedJac,
																		void* comm);
		virtual int evalHistoryStart(int historySegmentIndex,
																	double* p,
																	double& historyStart,
																	void* comm);
		virtual int getHistoryStartDiscDegree(int historySegmentIndex,
																	int& historyStartDiscDegree,
																	void* comm);
		int evalHistoryStartBase(int historySegmentIndex,
																	double* p,
																	double& historyStart,
																	ddemUserProvidedJac* userProvidedJac,
																	void* comm);
    virtual int evalDerivHistory(int historySegmentIndex,
																		double t,
                                    double* p,
                                    double* h,
																	  void* comm);
    virtual int evalHistory(int historySegmentIndex,
																		double t,
                                    double* p,
                                    double* h,
																	  void* comm);
    int evalHistoryBase(int historySegmentIndex,
																		double t,
                                    double* p,
                                    double* h,
                                    ddemUserProvidedJac* userProvidedJac,
																	  void* comm);
		virtual int compT0(double* p,
               double& t0,
							void* comm);
		int compT0Base(double* p,
               double& t0,
               ddemUserProvidedJac* userProvidedJac,
							void* comm);
		virtual int compT0DiscDegree(double* p,
               int& t0DiscDegree,
							void* comm);
		virtual int compY0(double* p,
               double* y0,
							 double* historyStarts,
							 void* comm);
		int compY0Base(double* p,
               double* y0,
							 double* historyStarts,
               ddemUserProvidedJac* userProvidedJac,
							 void* comm);
		int compZ0(double t0,double * y0,
													 double* p,
													 double* z0,
													 double* historyStarts,
													 void* comm);
		virtual int locateBetweenHistoryStarts(double t,
																						double* historyStarts,
																						int& betweenHisroryIndex);
		virtual int getContReductionFactor(){return 0;};
	
		virtual int getContPropagFactor(int delayArgumentIndex){
			if(delayArgumentIndex < nu_)
				return 1;
			else
				return 0;
		}
		virtual int stateDependent(){
			return stateDependent_;
		}
private:
	virtual void createUserProvidedJacs_();
	virtual void destroyUserProvidedJacs_();
};
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
#endif /* __DDEM_SYSTEM*/
