/* DOUBLE CHECKED = YES*/
#include <iostream>
using namespace std;

#include "ivp2dde_improved_CRK.h"

#include "ddem_trace.h"
#include "ddem_error_codes.h"
#include "ddem_basic_defins.h"
/*-------------------*/
extern "C" {
  extern int rkd56_(
int *n,
ddemIVPfcn,
double *x, double *y, double *xend, double *tol,
int *ind,
double *c__,
int *nw,
double *w,
int interp_Flag,
void* ivp2ddeCommVoid);
  extern int deriv_(
int *n,
double *x, double *y, double *xout,double *ypout,
double *htrial,
int *nw,
double *w,
int interp_Flag
);
  extern int intrp_(
int *n,
double *x, double *y, double *xout,double *yout,
double *htrial,
int *nw,
double *w,
int interp_Flag
);
}

/*-------------------*/
IVP2DDEImprovedCRK::IVP2DDEImprovedCRK(int interp_Flag)
{
	c_ = NULL;
	w_ = NULL;
	interp_Flag_ = interp_Flag;
	prepared_ = DDEM_FALSE;
}
/*-------------------*/
IVP2DDEImprovedCRK::~IVP2DDEImprovedCRK()
{
	if(w_ != NULL){ delete w_; w_ = NULL;};
	if(c_ != NULL){ delete c_; c_ = NULL;};
}
/*-------------------*/
int IVP2DDEImprovedCRK::getNInterploatorCoefficens()
{
	if( interp_Flag_ <= 1)
		return 6;
	else
		return 7;
}
/*-------------------*/
int IVP2DDEImprovedCRK::getRequiredContinuity()
{
	return 8;
}
/*-------------------*/
void IVP2DDEImprovedCRK::prepare(int nVariables,
																 double relTolerance,
																	double* absTolerance)
{
	if(	prepared_ == DDEM_TRUE){
		delete w_; w_ = NULL;
		delete c_; c_ = NULL;
	}
  n_ = nVariables;
	nw_ = n_;
	
	w_ = new double[nw_ * 30];

	c_ = new double[30 + n_];

	for(int i=0; i < 30 ; i++)
		c_[i] = 0;
	c_[9 - 1] = 1;
	c_[1 - 1] = 4;
	/*-------*/
	prepared_ = DDEM_TRUE;

	tol_ = relTolerance;

	for(int i = 0 ; i < n_; i++)
		c_[30  + i ] = absTolerance[i] / relTolerance;
}
/*-------------------*/
void IVP2DDEImprovedCRK::restart()
{
	ind_ = 2;

	c_[4 - 1] = c_[18 - 1];

}
/*-------------------*/
int IVP2DDEImprovedCRK::tryStep(double& t,double tEnd,
																					double* y,
																	ddemIVPfcn IVPfcn, void* ivp2ddeCommVoid)
{
	if(DDEM_TRACE_SIMULATOR)
		cout << "call to IVP2DDEImprovedCRK::tryStep()" << endl;
  double  xend = tEnd;
	
	rkd56_(&n_, IVPfcn, &t, y, &xend, &tol_, &ind_, c_, &nw_, w_, interp_Flag_, ivp2ddeCommVoid);


	if(ind_ == 3)
		return DDEM_AT_END;
	else if(ind_ == 5)
		return DDEM_ACCEPT;
	else if(ind_ == 6)
		return DDEM_REJECT;
	else if(ind_ < 0)
		return DDEM_ERR_IVP2DDE_BASE + ind_;
	else{
		cout << "IVP2DDEImprovedCRK::tryStep"<< endl <<
			"<<unexpected situation>>" << endl;
		return DDEM_ERR_UNEXPECTED;
	}
}
/*-------------------*/
void IVP2DDEImprovedCRK::storeStep(double& t, double* y, double** interpolant,
																		double& stepSize)
{
	t = c_[17-1];
	if(DDEM_TRACE_SIMULATOR)
		cout << "call to IVP2DDEImprovedCRK::storeStep() t = "<< t <<endl;
	stepSize = c_[18-1];
	for(int i=0; i < n_; i++)
		y[i] = w_[(13-1)*nw_ + i];
	/*-----------------------------------------------*/
	/* interp_Flag = 0,1						      	*/
	/*	y(i), w	1,13,9,10,11,12							*/
	/*					0  1 2  3  4 5							*/
	/* interp_Flag >= 2						        	*/
	/*	y(i), w	1,13,9,15,16,17,18					*/
	/*          0  1 2  3  4  5 6           */
	/*--------------------------------------*/
	/*	common  w 1,13,9 */
	int j;
	for(j=0; j < n_; j++){
		interpolant[j][0] = w_[(1-1)*nw_ + j];
		interpolant[j][1] = w_[(13-1)*nw_ + j];
		interpolant[j][2] = w_[(9-1)*nw_ + j];
	}
	if(interp_Flag_ <= 1){
	/*	  w 10,11,12		*/
		for(j=0; j < n_; j++){
			interpolant[j][3] = w_[(10-1)*nw_ + j];
			interpolant[j][4] = w_[(11-1)*nw_ + j];
			interpolant[j][5] = w_[(12-1)*nw_ + j];
		}
	}else{
	/*		w	15,16,17,18 */
		for(j=0; j < n_; j++){
			interpolant[j][3] = w_[(15-1)*nw_ + j];
			interpolant[j][4] = w_[(16-1)*nw_ + j];
			interpolant[j][5] = w_[(17-1)*nw_ + j];
			interpolant[j][6] = w_[(18-1)*nw_ + j];
		}
	}
}
/*-------------------*/
int IVP2DDEImprovedCRK::interpolate(int nVariables,
																		double tLeftMeshPoint,
																		 double* yLeftMeshPoint,
																		 double t, double* y, double stepSize,
																		 double** interpolant,
																		 int interpSelect)
{
  n_ = nVariables;
	nw_ = n_;
	
	double* wInterp;

	wInterp = new double[nw_ * 30];

	int j;
	for(j=0; j < n_; j++){
		wInterp[(1-1)*nw_ + j] = interpolant[j][0];
		wInterp[(13-1)*nw_ + j] = interpolant[j][1];
		wInterp[(9-1)*nw_ + j] = interpolant[j][2];
	}
	if(interp_Flag_ <= 1){
	/*	  w 10,11,12		*/
		for(j=0; j < n_; j++){
			wInterp[(10-1)*nw_ + j] = interpolant[j][3];
			wInterp[(11-1)*nw_ + j] = interpolant[j][4];
			wInterp[(12-1)*nw_ + j] = interpolant[j][5];
		}
	}else{
	/*		w	15,16,17,18 */
		for(j=0; j < n_; j++){
			wInterp[(15-1)*nw_ + j] = interpolant[j][3];
			wInterp[(16-1)*nw_ + j] = interpolant[j][4];
			wInterp[(17-1)*nw_ + j] = interpolant[j][5];
			wInterp[(18-1)*nw_ + j] = interpolant[j][6];
		}
	}
	/*-----------------*/
	if(interpSelect == DDEM_INTERP)
		intrp_(&n_, &tLeftMeshPoint, yLeftMeshPoint,
				 &t, y, &stepSize,
				 &nw_, wInterp, 
				 interp_Flag_); 
	else
		deriv_(&n_, &tLeftMeshPoint, yLeftMeshPoint,
				 &t, y, &stepSize,
				 &nw_, wInterp, 
				 interp_Flag_);
	delete wInterp;
	return DDEM_SUCCESS;
}
