#include <iostream>
using namespace std;
#include <fstream>
#include <iomanip>
#include <string>

#include <math.h>
#include <stdlib.h>
#include <time.h>

#include "ddem_system.h"
#include "ddem_simulator.h"
#include "ivp2dde_improved_CRK.h"

#include "ddem_parameter_estimator.h"

#include "SYSTEMS/systems.h"

/* only one of these is necessary */
#include "SYSTEMS/system_01.h"
#include "SYSTEMS/system_02.h"
#include "SYSTEMS/system_09.h"
#include "SYSTEMS/system_10.h"

#define my_rand(min, max) (int) (min + ((max) - (min)) * rand()/((double)RAND_MAX + 1))

void get_data_and_constraints_01(int nParameters, ddemData* data,
																 ddemConstraints* constraints);
void get_data_and_constraints_02(int nParameters, ddemData* data,
																 ddemConstraints* constraints);
void get_data_and_constraints_09(int nParameters, ddemData* data,
																 ddemConstraints* constraints);
void get_data_and_constraints_10(int nParameters, ddemData* data,
																 ddemConstraints* constraints);


int main(void)
{

	int experimentIdentifier = 4; /* used to select different experiments     */
	                              /* conveniently = 1,2,3,4                   */
																/*      corresponding systems are: 9,1,10,2 */

	int nRuns = 1;

	int solvers[] ={DDEM_PE_VERY_SIMPLE, DDEM_PE_SIMPLE, DDEM_PE_NEW, DDEM_PE_NEW_CE};
	/*                   0                   1              2            3          */
	int selectedSolver = 3;


	ddemSystem* system;
  system = new ddemSystem();

	ddemSimulationAttributes* simA;

	ddemData* data;
  data = new ddemData();

  ddemConstraints* constraints;
  constraints = new ddemConstraints();

	double TOL = 1e-6;
	/* create the system and set all required functions and parameters*/
	switch(experimentIdentifier){
		case 1:
			getSystem_09(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_09(TOL, simA);
			get_data_and_constraints_09(system->getNParameters(), data, constraints);
			break;
		case 2:
			getSystem_01(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_01(TOL, simA);
			get_data_and_constraints_01(system->getNParameters(), data, constraints);
			break;
		case 3:
			getSystem_10(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_10(TOL, simA);
			get_data_and_constraints_10(system->getNParameters(), data, constraints);
			break;
		case 4:
			getSystem_02(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_02(TOL, simA);
			get_data_and_constraints_02(system->getNParameters(), data, constraints);
			break;
		default:
			cout << "Invalid experimentIdentifier " << endl;
			return -1;
	}

	int retCode;

	retCode = system->checkIntegrity();


	if(retCode < 0){
		cout << "The DDE system is not well defined." << endl <<
			      "Error Code = " << retCode << endl;
		return retCode;
	}else
		cout << "The DDE system is well defined." << endl;
	/*-----------------------------------------------*/
	/*-----------------------------------------------*/
	IVP2DDEImprovedCRK* improvedCRK;
	improvedCRK = new IVP2DDEImprovedCRK(0/*interp_Flag*/);
	
	ddemSimulator*	simulator;
	simulator = new ddemSimulator(improvedCRK);
	
	/*-----------------------------------------------*/
	srand( (unsigned)time( NULL ) );
	
	double* perturbedOriginalParameters;
	perturbedOriginalParameters = new double[system->getNParameters()];
	double* optimumParameters;
	optimumParameters = new double[system->getNParameters()];
	/*-----------------------------------------------*/
	double timeTotal[4];
	int nFEvalsTotal[4];
	int nReferencedParameterValuesTotal[4];
	double finalObjectiveValueTotal[4];
	double parameterDiffNormTotal[4];
	for(int solverID = 0; solverID < 4; solverID++){
		timeTotal[solverID] = 0.0;
		nFEvalsTotal[solverID] = 0;
		nReferencedParameterValuesTotal[solverID] = 0;

		finalObjectiveValueTotal[solverID] = 0.0;
		parameterDiffNormTotal[solverID] = 0.0;
	}


	for(int runCounter = 1; runCounter <= nRuns; runCounter++){
		cout << "************************************* runCounter = " << 
			runCounter  << "**********************************"<< endl;
		for(int i=0; i < system->getNParameters(); i++){
			if(constraints->isFree(i) == DDEM_TRUE)
					perturbedOriginalParameters[i] = simA->parameters[i] * (1.0 + my_rand(-10,10)/100.0);
				else
					perturbedOriginalParameters[i] = simA->parameters[i];
				optimumParameters[i] = perturbedOriginalParameters[i];
		}
		for(int solverID = 0; solverID < 4; solverID++){ // run all the solvers
		//for(int solverID = selectedSolver; solverID < selectedSolver + 1; solverID++){
			cout << "-------------------------------" << "SOLVER ID = " << solverID <<
				"--------------------------------------"<< endl;
			ddemParameterEstimator* parameterEstimator;
			parameterEstimator = new ddemParameterEstimator();
			ddemParameterEstimationStatistics* statistics;
			statistics	= new	ddemParameterEstimationStatistics();
			
			clock_t start, end;
			double cpu_time_used;
			
			start = clock();/*TIME TIME TIME TIME*/
			int retCodeTemp =
			parameterEstimator->EstimateParameters(simulator, system, data,
																								constraints,
																								optimumParameters,
																								simA->relTolerance,
																								simA->absTolerance,
																								simA->vanishingConst,
																								statistics,
																								solvers[solverID],
																								simA->comm);
      end = clock();/*TIME TIME TIME TIME*/
			cpu_time_used = ((double) (end - start)) / CLOCKS_PER_SEC;
			timeTotal[solverID] += cpu_time_used;

			cout << "return code = " << retCodeTemp << "\n";

			cout << "nFEvals = " << statistics->nFEvals << endl;
			nFEvalsTotal[solverID] += statistics->nFEvals;
			cout << "nReferencedParameterValues = " << statistics->nReferencedParameterValues << endl;
			nReferencedParameterValuesTotal[solverID] += statistics->nReferencedParameterValues;

			cout << "Final Objective Value = " << statistics->finalObjectiveValue << endl;
			finalObjectiveValueTotal[solverID] += statistics->finalObjectiveValue;

			
			double parameterDiffNorm = 0.0;
			for(int i=0; i < system->getNParameters(); i++)
				parameterDiffNorm = ddem_max(parameterDiffNorm, fabs(perturbedOriginalParameters[i] - optimumParameters[i]));

			cout << "Maximum Parameter Difference = " << parameterDiffNorm << endl;
			parameterDiffNormTotal[solverID] += parameterDiffNorm;
			
			cout << endl;
			for(int i=0; i < system->getNParameters(); i++){
				cout << "originalParameters[" << i << "] = ";
				cout.setf(ios::fixed); cout << setprecision(16); 
				cout << simA->parameters[i] << "\n";
			}
			cout << endl;
			for(int i=0; i < system->getNParameters(); i++){
				cout << "perturbedOriginalParameters[" << i << "] = ";
				cout.setf(ios::fixed); cout << setprecision(16); 
				cout << perturbedOriginalParameters[i] << "\n";
			}
			cout << endl;
			for(int i=0; i < system->getNParameters(); i++){
				cout << "optimumParameters[" << i << "] = ";
				cout.setf(ios::fixed); cout << setprecision(16); 
				cout << optimumParameters[i] << "\n";
			}
			delete parameterEstimator;
			delete statistics;
		}/* solverID */
	}/* runCounter */
	for(int solverID = 0; solverID < 4; solverID++){ // run all the solvers
	//for(int solverID = selectedSolver; solverID < selectedSolver + 1; solverID++){
		cout << "-----------------------------------------" << "SOLVER ID = " << solverID <<
			"--------------------------------------"<< endl;

		cout << "Average Values in " << nRuns << " runs" << endl << endl;
		cout << "nFEvals = " << (int)(nFEvalsTotal[solverID] / (double)nRuns) << endl;
		cout << "nReferencedParameterValues = " << nReferencedParameterValuesTotal[solverID] /(double)nRuns << endl;

		cout << "Final Objective Value = " << finalObjectiveValueTotal[solverID] / (double)nRuns << endl;
		cout << "Time = " << timeTotal[solverID] / (double)nRuns << endl;
	}

	delete optimumParameters;
	delete perturbedOriginalParameters;
	delete simulator;
	delete improvedCRK;

	delete constraints;
	delete data;
	delete system;
	delete simA;

	return 0;
}

/************************************************************/
void get_data_and_constraints_01(int nParameters, ddemData* data,
												ddemConstraints* constraints)
{
	string dataFileName;

	dataFileName = "IN//example_01.d"; // change
	
	data->Load(dataFileName);

	int nLinearConstraints;
	int nNonlinearConstraints;
  double infinityBound;

	nLinearConstraints = 0;// change
	nNonlinearConstraints = 0; // change
  infinityBound = 1e+20; // change or leave 
	
	constraints->create(nParameters, nLinearConstraints, nNonlinearConstraints, infinityBound);


	constraints->setSimpleLowerBound(0, .01);
	constraints->setSimpleUpperBound(0, 1.9);
	// add statements for setting the user defined constraints
  // for example : 
	// constraints->setSimpleLowerBound(parameterIndex, value);
	// see "ddem_constraints.h" for the full list of commands
}
/************************************************************/
void get_data_and_constraints_02(int nParameters, ddemData* data,
												ddemConstraints* constraints)
{
	string dataFileName;

	dataFileName = "IN//example_02.d"; // change
	
	data->Load(dataFileName);

	int nLinearConstraints;
	int nNonlinearConstraints;
  double infinityBound;

	nLinearConstraints = 0;// change
	nNonlinearConstraints = 0; // change
  infinityBound = 1e+20; // change or leave 
	
	constraints->create(nParameters, nLinearConstraints, nNonlinearConstraints, infinityBound);
	constraints->setSimpleEquality(3, 33.0 / 100.0);
	constraints->setSimpleEquality(4, - 1.0 / 10.0);
	constraints->setSimpleEquality(5, 111.0 / 50.0);
	constraints->setSimpleEquality(6, 1.0 / 10.0);

	constraints->setSimpleLowerBound(0, .05);
	// add statements for setting the user defined constraints
  // for example : 
	// constraints->setSimpleLowerBound(parameterIndex, value);
	// see "ddem_constraints.h" for the full list of commands
}
/************************************************************/
void get_data_and_constraints_09(int nParameters, ddemData* data,
												ddemConstraints* constraints)
{
	string dataFileName;

	dataFileName = "IN//example_09.d";/* CHANGE */
	
	data->Load(dataFileName);

	int nLinearConstraints;
	int nNonlinearConstraints;
  double infinityBound;

	nLinearConstraints = 0;/* FILL */
	nNonlinearConstraints = 0; /* FILL */
  infinityBound = 1e+20; /* CHANGE */
	
	constraints->create(nParameters, nLinearConstraints, nNonlinearConstraints, infinityBound);
  
	constraints->setSimpleLowerBound(0, .05);
	constraints->setSimpleEquality(1, 8.0);

}
/************************************************************/
void get_data_and_constraints_10(int nParameters, ddemData* data,
												ddemConstraints* constraints)
{
	string dataFileName;

	dataFileName = "IN//example_10.d"; // change
	
	data->Load(dataFileName);

	int nLinearConstraints;
	int nNonlinearConstraints;
  double infinityBound;

	nLinearConstraints = 0;// change
	nNonlinearConstraints = 0; // change
  infinityBound = 1e+20; // change or leave 
	
	constraints->create(nParameters, nLinearConstraints, nNonlinearConstraints, infinityBound);

	constraints->setSimpleLowerBound(0, .05);
	constraints->setSimpleLowerBound(1, 5.0);
	// add statements for setting the user defined constraints
  // for example : 
	// constraints->setSimpleLowerBound(parameterIndex, value);
	// see "ddem_constraints.h" for the full list of commands
}
/************************************************************/
