#include <iostream>
using namespace std;

//#include <fstream>
#include <iomanip>
#include <string>

#include <math.h>

#include "ddem_system.h"
#include "ddem_simulator.h"
#include "ivp2dde_improved_CRK.h"
#include "ddem_sol_out.h"
#include "ddem_error_codes.h"

#include "SYSTEMS/systems.h"

/* only one of these is necessary */
#include "SYSTEMS/system_01.h"
#include "SYSTEMS/system_02.h"
#include "SYSTEMS/system_03.h"
#include "SYSTEMS/system_04.h"
#include "SYSTEMS/system_05.h"
#include "SYSTEMS/system_06.h"
#include "SYSTEMS/system_07.h"
#include "SYSTEMS/system_08.h"


#define round(x) ((int)floor((x) + 0.5))




int main(void)
{
	int experimentIdentifier = 2; /* used to select different experiments conveniently = 1,2,3,4,5,6,7,8 */
		
	ddemSystem* system;
  system = new ddemSystem();

	ddemSimulationAttributes* simA;
	

	double TOL = 1e-9;
	/* create the system and set all required functions and parameters*/
	switch(experimentIdentifier){
		case 1:
			getSystem_01(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_01(TOL, simA);
			break;
		case 2:
			getSystem_02(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_02(TOL, simA);
			break;
		case 3:
			getSystem_03(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_03(TOL, simA);
			break;
		case 4:
			getSystem_04(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_04(TOL, simA);
			break;
		case 5:
			getSystem_05(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_05(TOL, simA);
			break;
		case 6:
			getSystem_06(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_06(TOL, simA);
			break;
		case 7:
			getSystem_07(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_07(TOL, simA);
			break;
		case 8:
			getSystem_08(system);
			simA = new ddemSimulationAttributes(system->getNVariables(), system->getNParameters());
			getSimulationA_08(TOL, simA);
			break;
		default:
			cout << "Invalid experimentIdentifier " << endl;
			return -1;
	}

	int retCode;

	retCode = system->checkIntegrity();


	if(retCode != DDEM_TRUE){
		cout << "The DDE system is not well defined." << endl <<
			      "Error Code = " << retCode << endl;
		return retCode;
	}else
		cout << "The DDE system is well defined." << endl;



	IVP2DDEImprovedCRK* improvedCRK;
	improvedCRK = new IVP2DDEImprovedCRK(0/*interp_Flag*/);

	ddemSimulator*	simulator;
	simulator = new ddemSimulator(improvedCRK);

	ddemSimulationSolution*	simulationSolution;

	simulationSolution = new ddemSimulationSolution();
	retCode =
		simulator->simulate(system, simA->tEnd,
									simA->parameters,
									simA->relTolerance,
									simA->absTolerance,
									simA->vanishingConst,
									simA->comm,
									simulationSolution);
	
	cout << endl << endl << "Return code = " << retCode << endl;
	if(retCode != DDEM_SUCCESS)
		return retCode;

	/*00000000000000000000000000000000000000000000000000000000000000000*/
	int outYIndex = 0;/* The chosen component for output = 0,1,..., (nVariables - 1) */
	
	int interpSelect = DDEM_INTERP;/* DDEM_INTERP = value required, DDEM_DERIV_INTERP = derivative required*/

	string fileNameBase;
	string fileNameOutUniform;
	
	fileNameBase = "OUT//experiment_";
	fileNameBase.append(ddemInt2string(experimentIdentifier));
	fileNameBase.append("_");
	fileNameBase.append(ddemInt2string(round(-log10(TOL))));

	fileNameOutUniform = fileNameBase; 
	fileNameOutUniform.append("_U.dat");

	/* hight precision uniform output : suitable for error analysis, ... */
	ddemOutSolUniform(simulationSolution,
											outYIndex,
											interpSelect,
											1000/*nOutPoints*/,
											fileNameOutUniform,
											16  /*precision*/,
											simulator);
	
	//ddemOutSolUniformAll(simulationSolution,
	//										interpSelect,
	//										1000/*nOutPoints*/,
	//										fileNameOutUniform,
	//										16  /*precision*/,
	//										simulator);

	/* low precision non-uniform : suitable for plots*/
	string fileNameOutNonUniform;
	string fileNameOutNonUniformDisc;

	fileNameOutNonUniform = fileNameBase; 
	fileNameOutNonUniform.append("_NU.dat");

	fileNameOutNonUniformDisc = fileNameBase; 
	fileNameOutNonUniformDisc.append("_DISC.dat");

	//ddemOutSolNonUniform(simulationSolution,
	//										outYIndex,
	//										interpSelect,
	//										1000/*nOutPoints*/,
	//										fileNameOutNonUniform,
	//										fileNameOutNonUniformDisc,
	//										8  /*precision*/,
	//										simulator);
	ddemOutSolNonUniformAll(simulationSolution,
											interpSelect,
											1000/*nOutPoints*/,
											fileNameOutNonUniform,
											fileNameOutNonUniformDisc,
											8  /*precision*/,
											simulator);
	/*-------------- display integration statistics -----------*/
	cout << endl;
	cout << setiosflags( ios::scientific | ios::showpoint ) << setprecision(4); 
	cout << "TOL  = " << simA->relTolerance << endl;

	cout << endl;
	cout << "Number of successful steps = " << simulationSolution->statistics->nSteps << endl;
	cout << "Number of failed steps = " << simulationSolution->statistics->nFails << endl;
	cout << "Number of function(right hand side) evaluations = " << simulationSolution->statistics->nFEvals << endl;
	cout << "Number of vanishing steps = " << simulationSolution->statistics->nVanishingSteps << endl;
	cout << "Number of iteraions caused by vanishing steps = " << simulationSolution->statistics->nIteraions4VanishingSteps << endl;

	/* Display discontinuities along with their degree */
	cout << setiosflags( ios::scientific | ios::showpoint ) << setprecision(15); 
	cout << endl << "Tracked Discontinuities:" <<endl;
	for(int i=0; i < simulationSolution->nLambdas; i++)
		cout << "i = " << i << " lambda = " << 
		simulationSolution->meshPoints[simulationSolution->lambdasIndices[i]] <<
		" DiscDegree = " << simulationSolution->lambdasDiscDegrees[i] << endl;
	
	/* Evaluete and display the solution values at the end point*/
	double* y;
	int m = simulationSolution->getNVariables();
	y	= new double[m];
	double t = simulationSolution->meshPoints[simulationSolution->nMeshPoints - 1]/* end point */;
	simulator->evaluateSolution(t, y, simulationSolution, DDEM_INTERP, DDEM_UNKNOWN);
	cout << endl;
	cout << "Computed values at the endpoint: ";
	cout.unsetf(ios::scientific);
	cout.setf(ios::fixed);
	cout << setprecision(4); 
	cout << "t = " << t << endl;

	cout.unsetf(ios::fixed);
	cout << setiosflags( ios::scientific | ios::showpoint ) << setprecision(16); 
	for(int i=0; i < m; i++)
		cout << "y[" << i << "] = " << y[i] << endl;

	/* Use exact values to report errors */
	double* yExact;
	yExact = new double[m];
	if(experimentIdentifier == 1){
		double zeta2=4.0+2.0*log(2.0);
		yExact[0] = 4.0 - 2.0 * log(1.0+zeta2-t); 
	}else if(experimentIdentifier == 2){ 
		yExact[0] = 0.3318616184531945;
		yExact[1] = 2.2222766635175959;
	}else if(experimentIdentifier == 3){
		yExact[0] = pow(exp(1.0)/(3.0-log(t)),exp(1.0));
	}else if (experimentIdentifier == 5){
		yExact[0] = pow(t,3.0 + simA->parameters[0]);
	}else if(experimentIdentifier == 4){
		yExact[0] = 7357.6215803296436000;
	}else if(experimentIdentifier == 6){ 
		yExact[0] = 5.2312724900129703;
		yExact[1] = 0.0549084622861634;
		yExact[2] = 3.9851129367680183;
		yExact[3] = 5.9156352730635131;
	}else if(experimentIdentifier == 7){ 
		yExact[0] = 3.4997442903907818;
		yExact[1] = 9.9999175130290432;
		yExact[2] = 1.1993285802286043e2;
	}else if(experimentIdentifier == 8){ 
		yExact[0] = 8.4025502940268892e-1;
	}
	cout << endl;
	cout << "Absolute and relative errors for different components:" << endl;
	cout.unsetf(ios::fixed);
	cout << setiosflags( ios::scientific | ios::showpoint ) << setprecision(16); 
	for(int i=0; i < m; i++){
		double absError = fabs(y[i] - yExact[i]);
		double relError = absError / fabs(yExact[i]);
		cout << "absErr[" << i << "] = " << absError << "      " <<
			      "relErr[" << i << "] = " << relError << endl; 
	}
	delete yExact;
	/*------*/
	delete y; y = NULL;
	/*-------------------*/
	delete simulationSolution;
	delete simulator;
	delete improvedCRK;
	delete simA;
	delete system;
	/*-------------------*/
  return 0;
}
/************************************************************/